/******** <lalVerbatim file="BlockConditionDataCV"> ********
Author: Tibbits, M. M.
$Id: BlockConditionData.c,v 1.22 2003/07/06 20:06:08 jolien Exp $
********* </lalVerbatim> ********/

/******************************* <lalLaTeX file="BlockConditionDataC">
\subsection{Module \texttt{BlockConditionData.c}}
\label{s:BlockConditionData.c}

Routine that processes the input data object and repacks it into the
block parameter structure.  It then also sends out the first vector
to each slave node to be processed.

\subsubsection*{Prototypes}
\input{BlockConditionDataCP}
\input{BlockDistributeDataMasterCP}
\input{BlockDistributeDataSlaveCP}
\input{BlockParseDataCP}
\index{\texttt{LALConditionData()}}

\subsection*{Description}

Within the \texttt{LALConditionData} Routine, the block search master sends
the input object to the \texttt{BlockParseData} function.  This function
repacks the input in the masters parameter structure and stores the input
data in the parameter's data member.  The Master then sends the first vector
to each node to begin processing.  On each slave node, each slave simply
receives the first data vector and then exits.

\subsection*{Algorithms}

At this time, the \texttt{LALConditionData} routine does not manipulate
the data in any way.  It merely repacks the data in a LAL Datatype.

\subsection*{Uses}
 
\begin{itemize}
\item \texttt{BlockDistributeDataMaster}
\item \texttt{BlockDistributeDataSlave}
\item \texttt{LALMPISendMsg}
\item \texttt{LALMPIRecvMsg}
\item \texttt{LALMPISendINT4}
\item \texttt{LALMPIRecvINT4}
\item \texttt{LALMPISendREAL8}
\item \texttt{LALMPIRecvREAL8}
\item \texttt{LALDCreateVector}
\item \texttt{LALMPISendREAL8Vector}
\item \texttt{LALMPIRecvREAL8Vector}
\item \texttt{BlockParseData}
\item \texttt{LALDCreateVectorSequence}
\item \texttt{BlockBuildOutput} 
\item \texttt{LALDDestroyVector}
\end{itemize}

\subsection*{Notes}

\vfill{\footnotesize\input{BlockConditionDataCV}}

 *************************************************** </lalLaTeX> */

#include "Block.h"

NRCSID( BLOCKCONDITIONDATAC, "$Id: BlockConditionData.c,v 1.22 2003/07/06 20:06:08 jolien Exp $");

/************************ <lalVerbatim file="BlockConditionDataCP"> */
void LALConditionData
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams,
	LALMPIParams	*mpiParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*myParams;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "ConditionData", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Sanity Checks on Paramter Structure  */
	ASSERT(searchParams, status,BLOCKH_ENULL, BLOCKH_MSGENULL);

	/*  Variable Initialization  */
	myParams = ((BlockSearchParams*)(searchParams));

	/*  Get rank  */
	MPI_Comm_rank(*(mpiParams->comm), &(myParams->rank));

	/*  Start Master/Slave version  */
	/*  Which node is this running on?  */
	if (myParams->searchMaster)
	{
		/* master */
		BlockDistributeDataMaster(status->statusPtr, inout, searchParams, mpiParams );
	}
	else
	{
		/* slave */
		BlockDistributeDataSlave(status->statusPtr, inout, searchParams, mpiParams );
	}

	DETATCHSTATUSPTR( status );
	RETURN( status );
}


/*
 * Function BlockDistributeDataMaster
 *
 * Called from LALConditionData only on the search master node. It
 * uses the BlockParseData routine to repack the input and then it
 * sends the a data vector to each waiting slave.
 */

/************************ <lalVerbatim file="BlockDistributeDataMasterCP"> */
void BlockDistributeDataMaster
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams,
	LALMPIParams	*mpiParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*params;
	bandSegment		*tempPointer;
	bandObject		*bandSet;
	REAL8Vector		*tempVector;
	MPIMessage		mpiMsg;
	MPI_Comm		*comm;
	UINT4			currSlaves;
	UINT4			iterator;
	UINT4			level;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataMaster", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Check that input exists  */
	ASSERT(inout, status, BLOCKH_ENULL, BLOCKH_MSGENULL);
	ASSERT(inout->sequences, status, BLOCKH_ENULL, BLOCKH_MSGENULL);

	/*  Variable Initialization  */
	comm	= ((MPI_Comm*)(mpiParams->comm));
	params	= ((BlockSearchParams*)(searchParams));

	/* get the number of search slaves */
	MPI_Comm_size(*(mpiParams->comm), &(params->numSlaves));

	/* The number returned INCLUDES the search master, so subtract 1 */
	params->numSlaves -= 1;

	currSlaves	= params->numSlaves;

	/*  Repack data into usable LAL Structure  */
	BlockParseData ( status->statusPtr, inout, params );

	bandSet = params->data;

	/*  While there are still Slaves without Data, send more  */
	while(currSlaves > 0 )
	{
		REAL8  segmentLength;

		/*  Wait for Slaves to say, "I'm here"  */
		LALMPIRecvMsg( status->statusPtr, &mpiMsg, (*comm));
		CHECKSTATUSPTR( status );

		if( params->segmentsRemaining == 0 )
		{
			iterator = -1;

			/*  Send kill code  */
			LALMPISendINT4( status->statusPtr, &(iterator), mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );
			continue;
		}
		else
		{
			params->segmentsRemaining--;
		}

		params->times.startSec = (INT8) inout->sequences[0].range.dTime.startSec;
		params->times.startNan = (INT8) inout->sequences[0].range.dTime.startNan;
		params->times.stopSec = (INT8) inout->sequences[0].range.dTime.stopSec;
		params->times.stopNan = (INT8) inout->sequences[0].range.dTime.stopNan;
		params->times.timeStepSize = (INT8) inout->sequences[0].range.dTime.timeStepSize;
		params->times.numberSamples = (INT8) inout->sequences[0].range.dTime.numberSamples;


		/*  Send Data Paramters  */
		LALMPISendUINT4( status->statusPtr,&(params->times.startSec), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );
		LALMPISendUINT4( status->statusPtr,&(params->times.startNan), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );
		LALMPISendUINT4( status->statusPtr,&(params->times.stopSec), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );
		LALMPISendUINT4( status->statusPtr,&(params->times.stopNan), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );
		LALMPISendREAL8( status->statusPtr,&(params->times.timeStepSize), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );
		LALMPISendUINT8( status->statusPtr,&(params->times.numberSamples), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );


		/*  Send bandSet->segmentID */
		iterator = (params->numberOfSequences - params->segmentsRemaining) - 1;

		LALMPISendINT4( status->statusPtr, &(iterator), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send bandSet->levels  */
		LALMPISendINT2( status->statusPtr, &(bandSet->levels), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		segmentLength = ((REAL8)(bandSet->decomposition[0]->data->length)) / params->sampleRate[0];

		/*  Send decomposition[0]->data->length  */
		LALMPISendREAL8( status->statusPtr, &(segmentLength), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send startTime  */
		LALMPISendINT4( status->statusPtr, &(bandSet->decomposition[0]->startTime), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send Data in for loop  */
		for(level = 0; level < bandSet->levels; level++ )
		{
			tempPointer	= NULL;
			tempVector	= NULL;

			LALDCreateVector(status->statusPtr, &tempVector, bandSet->decomposition[level]->data->length);
			CHECKSTATUSPTR( status );

			for( iterator = 0; iterator < tempVector->length; iterator++ )
			{
				tempVector->data[iterator] = bandSet->decomposition[level]->data->data[iterator];
			}

			/*  Send varianceZero  */
			LALMPISendREAL8( status->statusPtr, &(params->varianceZero[level]), mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );

			/*  Send meanZero  */
			LALMPISendREAL8( status->statusPtr, &(params->meanZero[level]), mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );

			/*  Send Data to Slave  */
			LALMPISendREAL8Vector(status->statusPtr, tempVector, mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );

			/*  Clean up temporary Mem each time used  */
			LALDDestroyVector(status->statusPtr, &tempVector);
			CHECKSTATUSPTR( status );

			tempPointer = bandSet->decomposition[level];
			bandSet->decomposition[level] = ((bandSegment*)(bandSet->decomposition[level]->next));

			/*  Remove Sent BandSet Data  */
			LALDDestroyVector(status->statusPtr, &tempPointer->data);
			LALFree(tempPointer);
		}

		currSlaves--;
	}

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );
}

/*
 * Function BlockDistributeDataSlave
 *
 * Called from LALConditionData on each slave node. It
 * waits for the master to send the first data vector for
 * it to process and then it exits.
 */

/************************ <lalVerbatim file="BlockDistributeDataSlaveCP"> */
void BlockDistributeDataSlave
(
	LALStatus		*status,
	LALSearchInput	*inout,
	void			*searchParams,
	LALMPIParams	*mpiParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*params;
	bandSegment		*current;
	bandObject		*bandSet;
	MPIMessage		sendMsg;
	MPI_Comm		*comm;
	REAL8			length;
	INT4			startTime = 0;
	INT4			duration  = 0;
	INT4			level     = 0;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataSlave", BLOCKCONDITIONDATAC );
	ATTATCHSTATUSPTR (status);

	/*  Variable Initialization  */
	params	= ((BlockSearchParams*)(searchParams));
	comm	= ((MPI_Comm*)(mpiParams->comm));

	params->data = bandSet = (bandObject*)(LALMalloc( sizeof( bandObject)));
	if(!bandSet)
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	/*  Set up MPI Message  */
	sendMsg.source	= params->rank;
	sendMsg.msg	= PassData;
	sendMsg.send	= 1;

	/*  Send Message to say I am ready  */
	LALMPISendMsg( status->statusPtr, &sendMsg, 0, (*comm));
	CHECKSTATUSPTR( status );	

	/*  Receive Data Parameters  */
	LALMPIRecvUINT4( status->statusPtr,&(params->times.startSec), 0, (*comm));
	CHECKSTATUSPTR( status );
	LALMPIRecvUINT4( status->statusPtr,&(params->times.startNan), 0, (*comm));
	CHECKSTATUSPTR( status );
	LALMPIRecvUINT4( status->statusPtr,&(params->times.stopSec), 0, (*comm));
	CHECKSTATUSPTR( status );
	LALMPIRecvUINT4( status->statusPtr,&(params->times.stopNan), 0, (*comm));
	CHECKSTATUSPTR( status );
	LALMPIRecvREAL8( status->statusPtr,&(params->times.timeStepSize), 0, (*comm));
	CHECKSTATUSPTR( status );
	LALMPIRecvUINT8( status->statusPtr,&(params->times.numberSamples), 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Receive bandSet->segmentID */
	LALMPIRecvINT4( status->statusPtr, &(bandSet->segmentID), 0, (*comm));
	CHECKSTATUSPTR( status );

	if( bandSet->segmentID < 0 )
	{
		DETATCHSTATUSPTR( status );
		RETURN( status );
	}

	/*  Receive bandSet->levels  */
	LALMPIRecvINT2( status->statusPtr, &(bandSet->levels), 0, (*comm));
	CHECKSTATUSPTR( status );

	params->eventCounter = params->maxVectorLength * params->sampleRate[0] * bandSet->segmentID * bandSet->levels;

	bandSet->decomposition = (bandSegment**)(LALMalloc ( sizeof(bandSegment*) * bandSet->levels ));
	if( !bandSet->decomposition)
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	params->varianceZero = ((REAL8*)(LALMalloc ( sizeof(REAL8) * bandSet->levels )));
	if( !params->varianceZero )
	  {
	    ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	  }

	params->meanZero = ((REAL8*)(LALMalloc ( sizeof(REAL8) * bandSet->levels )));
	if( !params->meanZero )
	  {
	    ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	  }

	params->numberResults = 0;

	/*  Receive decomposition[0]->data->length  */
	LALMPIRecvREAL8( status->statusPtr, &(length), 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Receive startTime  */
	LALMPIRecvINT4( status->statusPtr, &(startTime), 0, (*comm));
	CHECKSTATUSPTR( status );

	for( level = 0; level < bandSet->levels; level++)
	{
		params->results = NULL;
		params->prevResults = NULL;
		bandSet->decomposition[level] = current = (bandSegment*)( LALMalloc( sizeof(bandSegment)));

		current->rhoN = current->index = current->value = 0;
		current->left	= NULL;
		current->right	= NULL;
		current->data	= NULL; 
		current->next	= NULL;

		current->startTime = startTime;
		current->duration = duration;
		current->level = level;

		LALDCreateVector( status->statusPtr, &(current->data), (length * params->sampleRate[level]));
		CHECKSTATUSPTR( status);

		/*  Receive varianceZero  */
		LALMPIRecvREAL8( status->statusPtr, &(params->varianceZero[level]), 0, (*comm));
		CHECKSTATUSPTR( status );

		/*  Receive meanZero  */
		LALMPIRecvREAL8( status->statusPtr, &(params->meanZero[level]), 0, (*comm));
		CHECKSTATUSPTR( status );

		/*  Slave receives data to process  */
		LALMPIRecvREAL8Vector(status->statusPtr, current->data, 0, (*comm));
		CHECKSTATUSPTR( status );
	}

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );
}

/************************ <lalVerbatim file="BlockParseDataCP"> */
void BlockParseData
(
	LALStatus	*status,
	LALSearchInput	*inout,
	void		*searchParams
)
/*********************************************** </lalVerbatim> */
{

  /*  Variable Declarations  */
  BlockSearchParams	*params;
  bandSegment		*current;
  bandSegment		*temp;
  bandObject		*bandSet;
  INT4			lengthMinusOverlap;
  INT4			vectorLength;
  INT4			iterator;
  INT4			offset;
  INT4			level;
  INT4			sampleRate;
  INT4			rescale;
  INT4			length;
  INT4			index;

  /*  Initialize status to pass to other functions  */
  INITSTATUS( status, "BlockParseData", BLOCKCONDITIONDATAC );
  ATTATCHSTATUSPTR (status);

  /*  Variable Initialization  */
  params	= ((BlockSearchParams*)(searchParams));
  rescale	= 1;
  temp 		= NULL;

  /*  Initialize the array of Pointers  */
  bandSet	= (bandObject*)(LALMalloc( sizeof(bandObject) ));
  if( !bandSet )
    {
      ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
    }

  bandSet->levels = inout->numberSequences;

  bandSet->decomposition = (bandSegment**)(LALMalloc ( sizeof(bandSegment*) * bandSet->levels ));
  if( !bandSet->decomposition )
    {
      ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
    }

  params->varianceZero = ((REAL8*)(LALMalloc ( sizeof(REAL8) * bandSet->levels )));
  if( !params->varianceZero )
    {
      ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
    }


  params->meanZero = ((REAL8*)(LALMalloc ( sizeof(REAL8) * bandSet->levels )));
  if( !params->meanZero )
    {
      ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
    }


  /*  Outer loop steps through bandSet levels  */
  /*  Middle loop steps through bandSet segments  */
  /*  Inner loop copies bandSet data  */

  for( level = 0; level < bandSet->levels; level++)
    {
      /*  Allow for simple overlapping  */
      if(params->overlapMethod)
	{
	  lengthMinusOverlap = params->maxVectorLength * params->sampleRate[level] - params->overlap;
	}
      else
	{
	  lengthMinusOverlap = params->maxVectorLength * params->sampleRate[level];
	}

      sampleRate = lengthMinusOverlap;

      /*  Process the entire band on Master to return Event 0  */
      /*  Allocate Band Segment for entire band  */
      temp = (bandSegment*)( LALMalloc( sizeof(bandSegment)));
      if( !temp )
	{
	  ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

      /*  Initialize band Segment  */
      temp->rhoN	 = temp->index = temp->value = 0;
      temp->left	 = NULL;
      temp->right	 = NULL;
      temp->data	 = NULL;

      /*  Allocate Data  */
      LALDCreateVector( status->statusPtr, &(temp->data), inout->sequences[level].dimensions[0]);
      CHECKSTATUSPTR( status );

      /*  Copy Input Data  */
      for( index = 0; index < inout->sequences[level].dimensions[0]; index++ )
	{
	  temp->data->data[index] = (REAL8)(inout->sequences[level].data.real4[index]);
	}

      /*  Find Mean & Variance  */
      LALDMoment(status->statusPtr,&(params->varianceZero[level]), temp->data, 2 );
      LALDMoment(status->statusPtr, &(params->meanZero[level]), temp->data, 1);

      LALDDestroyVector( status->statusPtr, &(temp->data));
      LALFree(temp);

      /*  Determine Number of Input Segments  */
      if( inout->sequences[level].dimensions[0] < params->maxVectorLength * params->sampleRate[level] )
	{
	  ABORT(status, BLOCKH_EDATA, BLOCKH_MSGEDATA);
	}

      /*  Measure for this level only  */
      params->numberOfSequences = 1;

      vectorLength = inout->sequences[level].dimensions[0] -  params->maxVectorLength * params->sampleRate[level];

      while( vectorLength > params->minVectorLength )
	{
	  params->numberOfSequences++;
	  vectorLength -= sampleRate;
	}

      params->results = NULL;
      params->prevResults = NULL;
      bandSet->decomposition[level] = current = (bandSegment*)( LALMalloc( sizeof(bandSegment)));


#ifdef BLOCKSEARCH__DEBUG
      fprintf(stdout,"MSG: inout->sequences[%d].dimensions[0] = %d\n",level,inout->sequences[level].dimensions[0]);
      fflush(stdout);
#endif


      for( iterator = 0; iterator < params->numberOfSequences - 1; iterator++)
	{
	  offset = iterator * sampleRate;

	  vectorLength = ((INT4)(params->maxVectorLength * params->sampleRate[level]));

	  current->rhoN	 = current->index = current->value = 0;
	  current->duration	= 0;

	  current->level = level;

	  current->left	 = NULL;
	  current->right = NULL;
	  current->data	 = NULL;

	  LALDCreateVector( status->statusPtr, &(current->data), vectorLength);
	  CHECKSTATUSPTR( status );

	  for( index = 0; index < vectorLength; index++ )
	    {
	      current->data->data[index] = (REAL8)(inout->sequences[level].data.real4[index + offset]);
	    }

	  /*  Must correct for metadata  */
	  current->startTime	= offset;

	  current->next = LALMalloc(sizeof(bandSegment));
	  if( !current->next )
	    {
	      ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	    }

	  current = ((bandSegment*)(current->next));

	  if( inout->sequences[level].dimensions[0] - (offset + sampleRate) < sampleRate )
	    break;

#ifdef BLOCKSEARCH__DEBUG
      fprintf(stdout,"MSG: Start of Sequnce[%d] = %d (%d points long)\n",iterator,offset,vectorLength);
      fflush(stdout);
#endif
	}

      offset = (params->numberOfSequences - 1)*sampleRate;

      current->rhoN = current->index = current->value = 0;
      current->duration = 0;

      current->level	= level;

      current->left	= NULL;
      current->right	= NULL;
      current->data	= NULL;
      current->next	= NULL;

      /*  Must correct for metadata  */
      current->startTime = offset;

      length = inout->sequences[level].dimensions[0] - offset;


#ifdef BLOCKSEARCH__DEBUG
      printf("Input bandSet decompositions = %d\n",bandSet->levels);
      printf("params->max = %d  total / max = %d\n", params->maxVectorLength, params->numberOfSequences);
      printf("Length of last vector := %d \n",length);
#endif


      current->data = NULL;
      LALDCreateVector( status->statusPtr, &(current->data), length);
      CHECKSTATUSPTR( status );

      for( index = offset; index < inout->sequences[level].dimensions[0]; index++ )
	{
	  current->data->data[index - offset]=(REAL8)(inout->sequences[level].data.real4[index]);
	}
    }

  params->segmentsRemaining = params->numberOfSequences;

  /*  Assert Usable amount of Data  */
  if ( params->numberOfSequences < 1 )
    {
      ABORT(status, BLOCKH_EDATA, BLOCKH_MSGEDATA);
    }

  /*  Global metadata  */
  params->times.startSec = (INT8) inout->sequences[0].range.dTime.startSec;
  params->times.startNan = (INT8) inout->sequences[0].range.dTime.startNan;
  params->times.stopSec = (INT8) inout->sequences[0].range.dTime.stopSec;
  params->times.stopNan = (INT8) inout->sequences[0].range.dTime.stopNan;
  params->times.timeStepSize = (INT8) inout->sequences[0].range.dTime.timeStepSize;
  params->times.numberSamples = (INT8) inout->sequences[0].range.dTime.numberSamples;

  /*  Store newly created bandSet object in master's params structure  */
  params->data = bandSet;

  /*  Normal Exit  */
  DETATCHSTATUSPTR (status);
  RETURN(status);
}
