/* -*- mode: c; c-basic-offset: 2; -*- */
/******** <lalVerbatim file="ExtTrigApplySearchCV"> ********
Author: Rahkola, R
$Id: ExtTrigApplySearch.c,v 1.13 2003/07/03 18:30:56 jolien Exp $
********* </lalVerbatim> ********/

/* Include I/O functions if we're in the debug stage */
#if defined (DEBUG_EXTTRIG) || defined (DUMMY_CODE)
#include <stdio.h>
#endif /*  defined (DEBUG_EXTTRIG) || defined (DUMMY_CODE)*/

/* C functions */
#include <math.h>
#include <mpi.h>            /* Contains all MPI_* functions */

/* standard LAL functions*/
#include <lal/LALStdlib.h>
#include <lal/AVFactories.h>
#include <lal/Date.h>
#include <lal/StochasticCrossCorrelation.h>
#include <lal/Comm.h>       /* LAL MPI interface */

/* Associated headers */
#include <ExtTrig.h>
#include <ExtTrigApplySearch.h>

#ifdef __cplusplus
extern "C" {
#endif

NRCSID( EXTTRIGAPPLYSEARCHC, "$Id: ExtTrigApplySearch.c,v 1.13 2003/07/03 18:30:56 jolien Exp $");

/********************* local static function declarations ********************/

static void
ApplySearchMaster(
		  LALStatus            *const status,
		  LALSearchOutput      *const output,
		  const LALSearchInput *const input,
		  ETSearchParams       *const params,
		  MPI_Comm              const comm   );

static void
ApplySearchSlave(
		 LALStatus            *const status,
		 LALSearchOutput      *const output,
		 const LALSearchInput *const input,
		 ETSearchParams       *const params,
		 MPI_Comm                    comm   );

static void
ExtTrigVariance(
		LALStatus             *const status,
		REAL4                 *const var1,
		REAL4                 *const var2,
		ETDataSegment         *const data   );

static void
ExtTrigCrossCorrelate(
		      LALStatus                     *const status,
		      ETCrossCorrOutput             *const ccoutput,
		      const ETDataSegment           *const ccinput,
		      const BOOLEAN                  makeSpectrum   );

#if DEBUG_EXTTRIG > 0
static void
PrintFSeries(
	     LALStatus                     *const status,
	     const COMPLEX8FrequencySeries *const fseries,
	     const UINT4                    binning      );
#endif /*  DEBUG_EXTTRIG > 0*/

/************************* wrapperAPI function definitions  ******************/

/*
 * Function LALApplySearch
 *
 * Determines if this node is a master or slave and calls the
 * appropriate function.
 */
void
LALApplySearch(
    LALStatus             *status,
    LALSearchOutput       *output,
    LALSearchInput        *input,
    LALApplySearchParams  *applyParams)
{
  /* cast the ETSearch parameter structure */
  ETSearchParams* const params = (ETSearchParams *) applyParams->searchParams;

  INITSTATUS( status, "LALApplySearch", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "exttrig:LALApplySearch\n" );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

  /*
   * Check that output and parameters are configured correctly
   *
   */
  ASSERT( output, status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL );
  ASSERT( !(output->result), 
	  status, APPLYSEARCHH_ENNUL, APPLYSEARCHH_MSGENNUL );
  ASSERT( applyParams, status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL );
  ASSERT( applyParams->mpiParams,
	  status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL );
  ASSERT( applyParams->searchParams,
	  status, APPLYSEARCHH_ENULL, APPLYSEARCHH_MSGENULL );

  /*
   * Determine whether we're running as master or as a slave
   *
   */
  MPI_Comm_rank( *(applyParams->mpiParams->comm), &(params->rank) );

  if ( params->searchMaster ) /* master */
  {
    /* check that rank number confirms whether we're searchMaster */
    ASSERT( params->rank == 0,
	    status, APPLYSEARCHH_ERANK, APPLYSEARCHH_MSGERANK );
    ApplySearchMaster( status->statusPtr, output, input, params,
		       *(applyParams->mpiParams->comm)          );
  }
  else                        /* slave */
  {
    /* check that rank number confirms whether we're searchSlave */
    ASSERT( params->rank > 0,
	    status, APPLYSEARCHH_ERANK, APPLYSEARCHH_MSGERANK );
    ApplySearchSlave( status->statusPtr, output, input, params,
		      *(applyParams->mpiParams->comm)          );
  }
  CHECKSTATUSPTR( status );
  
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

/*
 * Function ApplySearchMaster:  
 * Called from LALApplySearch only on the search master node. It's main job is
 * to wait for responses for each slave ie. retrieve the statistics and wait
 * for the slaves to end, then make the output.  Since this module will run
 * continuously during the search, we wish to maximize the frequency for which
 * it checks if a search slave is finished and ready for more data. To this end
 * we need the following functionality:
 * 0a) If not initialized
 *    - initialize ApplySearchMaster
 * 1a) If all data is finished, and any slaves have finished
 *    - build the LALSearchOutput structure
 *    - FINISH
 * 1b) If not finished
 *    - keep output->notFinished set to TRUE
 * 2a) If no slaves
 *    - cross-correlate chunk 'n' of data (= next unprocessed chunk)
 *    - update SegsToBeAnalyzed
 * 2b) If slave(s) exist and are active
 *    - wait for slave message
 *  2b.a) If slave needs data
 *    - send SegsToBeAnalyzed
 *    - send data to slaves
 *    - update SegsToBeAnalyzed
 *    - FINISH
 *  2b.b) If slave has results
 *    - get results for chunk 'n'
 * 3) Plug results for chunk 'n' into ETResult
 * 4) Update SegsAnalyzed
 * 5) Update search status
 * 6) FINISH
 */
static void
ApplySearchMaster(
		  LALStatus            *const status,
		  LALSearchOutput      *const output,
		  const LALSearchInput *const input,
		  ETSearchParams       *const params,
		  MPI_Comm              const comm   )
{
  /*UINT4                    i;       iterator(s)                       */
  MPIMessage               mpiMsg;      /* send/rec'v messages w/slave       */
  UINT4                    curSegment = /* index of first unanalyzed segment */
    params->data->filled - params->data->segsToBeAnalyzed;
  REAL4                    tmpVar1, tmpVar2;   /* calculated variance        */
  REAL8                    tmpTime1,tmpTime2;  /* used for getting seg delay */
  ETDataSegment            tmpSegment;  /* used for auto-correlation         */
  ETCrossCorrOutput        ccOutput;    /* used for auto-/cross-correlation  */
  BOOLEAN                  onsrc;
  BOOLEAN                  makeSpectrum =     /* if c.c. spectrum is used    */
    (params->initParams->initMask & 48);
  INT2Vector              *tmpLALUnit = NULL;  /* temp storage for LALUnit   */
  /* pointers to various useful stuff */
  ETSegment               *segments = params->segments->data;
  ETResult                *results = params->result;
  ETDataSegment           *dataSeg = params->data->data;

  INITSTATUS( status, "ApplySearchMaster", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

#ifdef DEBUG_EXTTRIG
  fprintf( stdout, "exttrig:LALApplySearch - %s\n", params->id );
  fflush ( stdout );
#endif /*  DEBUG_EXTTRIG*/

  /*
   * 0a) If not initialized
   *    - initialize ApplySearchMaster
   *      + allocate memory for ETResults structure (variance1, variance2,
   *        xCorrValue, xCorrDelay, xCorrLag )
   *      + debug initialization statement
   *      + allocate memory for xCorrSpectra (if no slaves)
   */
  if (params->initialised == 0)
    {
      MPI_Comm_size(comm, &(params->numSlaves)); /* get the number of nodes */
      params->numSlaves -= 1;     /* subtract 1 to get the number of slaves */
      params->curSlaves  = params->numSlaves;  /* start out with all slaves */

      /* Allocate results vectors with only as many segments as we can find */
      results->length     = params->data->segsToBeAnalyzed;
      results->variance1  = (REAL4 *)LALCalloc(results->length, sizeof(REAL4));
      results->variance2  = (REAL4 *)LALCalloc(results->length, sizeof(REAL4));
      results->xCorrValue = (REAL4 *)LALCalloc(results->length, sizeof(REAL4));
      results->xCorrDelay = (REAL4 *)LALCalloc(results->length, sizeof(REAL4));
      results->xCorrLag = (REAL4 *)LALCalloc(results->length, sizeof(REAL4));

#if DEBUG_EXTTRIG > 0
      /* a prefix to identify debugging messages */
      snprintf( params->id, IDLEN, "master[node %2d]", params->rank );
      fprintf( stdout, "%s: Current number of slaves: %u\n",
	       params->id, params->numSlaves );
      fflush( stdout );
#endif /*  #if DEBUG_EXTTRIG > 0*/

      /* Allocate results c.c. spectra if there aren't any slaves */
      if ( makeSpectrum && !(params->numSlaves) )
	results->xCorrSpectra = (COMPLEX8FrequencySeries **)
	  LALCalloc(params->data->filled, sizeof(COMPLEX8FrequencySeries *));

      params->initialised = 1;          /* ApplySearchMaster is initialized */

      output->numOutput   = 0;
      output->result      = NULL;
      output->fracRemaining = 1.0;
      output->notFinished = 1;
    }
  
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, 
    "%s: Current segment = %d, segsToBeAnalyzed = %d, segsAnalyzed = %d\n",
	   params->id, curSegment, params->data->segsToBeAnalyzed,
	   params->data->segsAnalyzed );
  fprintf( stdout, "%s: Number of active slaves = %d\n", params->id,
	   params->curSlaves );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

  /*
   * 1a) If all data is finished, and any slaves have finished
   *    - build the LALSearchOutput structure
   *    - FINISH
   */
  if ( params->curSlaves == 0 && params->data->segsToBeAnalyzed == 0 )
    {
      /* assign the SearchSummaryInfo number of events (analyzed segments) */
      params->info->numEvents = params->data->segsAnalyzed;

#if DEBUG_EXTTRIG < 3
      output->numOutput = 2;
      /* increase number of outPuts if we have spectra to produce */
      if ( makeSpectrum && !(params->numSlaves) )
	output->numOutput += params->result->filled;

      /* Here we'll allocate the results, thus it's not necessary to
       * (re-)allocate memory in the ExtTrigBuild* functions.
       */
      output->result = (outPut *) LALCalloc(output->numOutput, sizeof(outPut));
      
      /* build output database entries */
      ExtTrigBuildDB( status->statusPtr, output, params );
      CHECKSTATUSPTR( status );

      if ( makeSpectrum && !(params->numSlaves) ) {
	/* build output spectra if needed */
	ExtTrigBuildILWD( status->statusPtr, output, params );
	CHECKSTATUSPTR( status );
      }
#else
      fprintf( stdout, "%s: finishing ApplySearch loop\n", params->id );
      fflush( stdout );
#endif /*  DEBUG_EXTTRIG < 3*/

      output->fracRemaining = 0.0;
      output->notFinished   = 0;

      DETATCHSTATUSPTR( status );
      RETURN( status );
    }

  /*
   * 1b) If not finished
   *    - keep output->notFinished set to TRUE
   */
  else
    {
      output->notFinished   = 1;
      output->fracRemaining = (REAL4) (params->data->segsToBeAnalyzed / 
				       (REAL4) params->data->length    );
      output->numOutput     = 0;
    }

  /*
   * 2a) If no slaves
   *    - check whether c.c. spectra are expected for chunk 'n'
   *    - auto-correlate chunk 'n' of data (= next unprocessed chunk)
   *    - cross-correlate chunk 'n' of data (= next unprocessed chunk)
   *    - update SegsToBeAnalyzed
   */
  if ( !params->numSlaves )
    {
      /* check if output spectra are expected; if so, need to create spectra */
      onsrc = (curSegment == 0) ? 1 : 0;
      makeSpectrum = ((params->initParams->initMask & 16) && onsrc) ||
	((params->initParams->initMask & 32) && !onsrc);
      if (makeSpectrum) {
	ccOutput.ccSpectrum = results->xCorrSpectra[results->filled] =
	  (COMPLEX8FrequencySeries *)
	  LALMalloc( sizeof(COMPLEX8FrequencySeries) );
	LALCCreateVector( status->statusPtr, &(ccOutput.ccSpectrum->data),
			  dataSeg[curSegment].filter->data->length);
	CHECKSTATUSPTR( status );
	results->filled++;
      }


      /* Auto-correlation: TimeSeries #1 */
      tmpSegment.tsChan1   = dataSeg[curSegment].tsChan1;
      tmpSegment.tsChan2   = dataSeg[curSegment].tsChan1;
      tmpSegment.filter    = dataSeg[curSegment].filter;
      tmpSegment.fftPlan   = params->fftParams;
      tmpSegment.segmentID = dataSeg[curSegment].segmentID;
      ExtTrigCrossCorrelate( status->statusPtr, &ccOutput, &tmpSegment, 0 );
      CHECKSTATUSPTR( status );
      tmpVar1 = ccOutput.ccStatistic.value;

      /* Auto-correlation: TimeSeries #2 */
      tmpSegment.tsChan1   = dataSeg[curSegment].tsChan2;
      tmpSegment.tsChan2   = dataSeg[curSegment].tsChan2;
      ExtTrigCrossCorrelate( status->statusPtr, &ccOutput, &tmpSegment, 0 );
      CHECKSTATUSPTR( status );
      tmpVar2 = ccOutput.ccStatistic.value;

      /* Cross-correlation */
      ExtTrigCrossCorrelate( status->statusPtr, &ccOutput,
			     &(dataSeg[curSegment]), makeSpectrum );
      CHECKSTATUSPTR( status );

      params->data->segsToBeAnalyzed--;

#ifdef DUMMY_CODE
      if ( ! params->data->segsToBeAnalyzed ) {
	fprintf( stdout, "%s: TERMINATE ApplySearch loop\n", params->id );
	fflush ( stdout );
      }
#endif /*  DUMMY_CODE*/
    }

  /*
   * 2b) If slave(s) exist and are active
   *    - wait for slave message
   */
  else
    {
      LALMPIRecvMsg(status->statusPtr, &mpiMsg, comm);
      CHECKSTATUSPTR( status );

      /*
       *  2b.a) If slave needs data
       *    - send dataSegment ID to slaves
       *    - update SegsToBeAnalyzed
       *    - FINISH
       */
      if ( ! mpiMsg.send )
	{
	  /* update curSlaves */
	  if ( params->data->segsToBeAnalyzed == 0 )
	    {
	      curSegment++;
	      LALMPISendUINT4( status->statusPtr, &curSegment,
			       mpiMsg.source, comm            );
	      CHECKSTATUSPTR( status );

#ifdef DUMMY_CODE
	      fprintf( stdout,"%s: sent TERMINATE signal to slave[node %2d]\n",
		       params->id, mpiMsg.source );
	      fflush( stdout );
#endif /*  DUMMY_CODE*/

	      params->curSlaves--;
	    }
	  /* only send data if segsToBeAnalyzed > 0 */
	  else 
	    {
	      LALMPISendUINT4( status->statusPtr, &curSegment,
			       mpiMsg.source, comm            );
	      CHECKSTATUSPTR( status );

	      params->data->segsToBeAnalyzed--;
#if DEBUG_EXTTRIG > 1
	      fprintf( stdout, "%s: sent data to slave[node %2d]\n", 
		       params->id, mpiMsg.source );
	      fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
	    }

	  DETATCHSTATUSPTR( status );
	  RETURN( status );
	}
      /*
       *  2b.b) If slave has results
       *    - get results for chunk 'n'
       */
      else if ( (mpiMsg.send == 1) )
	{
	  /* create temporary storage for LALUnit */
	  LALI2CreateVector( status->statusPtr, &tmpLALUnit, 
			     (1 + 2 * LALNumUnits)          );
	  CHECKSTATUSPTR( status );

	  LALMPIRecvUINT4( status->statusPtr, &curSegment, 
			   mpiMsg.source, comm            );
	  CHECKSTATUSPTR( status );

	  LALMPIRecvREAL4( status->statusPtr, &(ccOutput.ccStatistic.value),
			   mpiMsg.source, comm             );
	  CHECKSTATUSPTR( status );

	  LALMPIRecvINT2Vector( status->statusPtr, tmpLALUnit,
				mpiMsg.source, comm           );
	  CHECKSTATUSPTR( status );
	  
	  LALMPIRecvREAL4( status->statusPtr, &(tmpVar1),
			   mpiMsg.source, comm             );
	  CHECKSTATUSPTR( status );

	  LALMPIRecvREAL4( status->statusPtr, &(tmpVar2),
			   mpiMsg.source, comm             );
	  CHECKSTATUSPTR( status );

	  ccOutput.ccStatistic.units.powerOfTen = tmpLALUnit->data[0];
	  memcpy( ccOutput.ccStatistic.units.unitNumerator, 
		  &(tmpLALUnit->data[1]), 
		  LALNumUnits * sizeof(INT2)      );
	  memcpy( ccOutput.ccStatistic.units.unitDenominatorMinusOne, 
		  &(tmpLALUnit->data[1 + LALNumUnits]), 
		  LALNumUnits * sizeof(INT2)                );

	  /* destroy temporary storage for LALUnit */
	  LALI2DestroyVector( status->statusPtr, &tmpLALUnit );
	  CHECKSTATUSPTR( status );

#if DEBUG_EXTTRIG > 1
	      fprintf( stdout, "%s: received data from slave[node %2d]\n", 
		       params->id, mpiMsg.source );
	      fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
	}
      else  /* Invalid MPI request from slave */
	{
	  ABORT( status, APPLYSEARCHH_ESEND, APPLYSEARCHH_MSGESEND );
	}
    }  /*  if ( params->curSlaves ) ...*/

  /*
   * 3) Plug results for chunk 'n' into ETResult
   */
  results->variance1[curSegment] = tmpVar1;
  results->variance2[curSegment] = tmpVar2;
  results->xCorrValue[curSegment] = ccOutput.ccStatistic.value;
  results->xCorrLag[curSegment] =
    segments[dataSeg[curSegment].segmentID].delay;
  LALGPStoFloat( status->statusPtr, &tmpTime1,
		 &(params->initParams->grbStart) );
  LALGPStoFloat( status->statusPtr, &tmpTime2,
		 &(segments[dataSeg[curSegment].segmentID].startGPS) );
  results->xCorrDelay[curSegment] = (REAL4)(tmpTime2 - tmpTime1);
  memcpy( &(results->xCorrUnit),
	  &(ccOutput.ccStatistic.units), sizeof(LALUnit) );

#if DEBUG_EXTTRIG > 1
  fprintf( stdout, 
   "%s: dataSeg[%2d](=>seg%2d): value %8.3g,\tlag %8.6f,\tdelay %8.6f\n",
	   params->id,
	   curSegment, dataSeg[curSegment].segmentID, 
	   results->xCorrValue[curSegment],
	   results->xCorrLag[curSegment],
	   results->xCorrDelay[curSegment] );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

  /*
   * 4) Update SegsAnalyzed
   */
  params->data->segsAnalyzed++;
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: segsAnalyzed = %2d\n", params->id, 
	   params->data->segsAnalyzed                   );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

  /*
   * 5) Update search status
   */
  output->fracRemaining = 1.0 - 
    (REAL4) ( params->data->segsAnalyzed / (REAL4) params->data->length );

  /*
   * 6)  FINISH
   */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

/*
 * Function ApplySearchSlave
 *
 * Called by LALApplySearch only on slave nodes. Each slave node
 * calculates one stochastic cross-correlation at a time. Since there may be
 * no slave nodes in the final configuration, we should put most of the search
 * code into another function and use ApplySearchSlave as a wrapper for that
 * function.  To this end, we need the following functionality:
 * 0a) If not initialized
 *    - initialize ApplySearchSlave (debugging purposes)
 * 1)  Get data from master
 * 1a) If segsToBeAnalyzed == 0
 *    - update notFinished
 * 1b) If segsToBeAnalyzed > 0
 *    - set the current segment
 * 2)  Perform cross-correlation for chunk 'n' of data
 * 2a) Calculate variance of the two time series
 * 2b) Calculate cross-correlation statistic for chunk 'n' of data
 * 3)  Send results to master
 * 4)  FINISH
 */
static void
ApplySearchSlave(
		 LALStatus            *const status,
		 LALSearchOutput      *const output,
		 const LALSearchInput *const input,
		 ETSearchParams       *const params,
		 MPI_Comm                    comm   )
{
  /*UINT4                    i;              * iterator                    */
  UINT4                    curSegment;  /* index of segment being analyzed   */
  MPIMessage               mpiMsg;      /* send/rec'v messages w/slave       */
  REAL4                    tmpVar1, tmpVar2;  /* calculated variance         */
  ETDataSegment            tmpSegment;  /* used for auto-correlation         */
  ETCrossCorrOutput        ccOutput;    /* used for auto-/cross-correlation  */
  INT2Vector              *tmpLALUnit = NULL; /* temp storage for LALUnit    */
  BOOLEAN                  onsrc;             /* if c.c. is for on-source    */
  BOOLEAN                  makeSpectrum =     /* if c.c. spectrum is saved   */
    (params->initParams->initMask & 48);
  /* pointers to various useful stuff */
  ETResult                *results = params->result;
  ETDataSegment           *dataSeg = params->data->data;

  INITSTATUS( status, "ApplySearchSlave", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

  /*
   * 0a) If not initialized
   *    - initialize ApplySearchSlave
   *      + allocate all memory for params->result->xCorrSpectra
   *        (but not for each pointer)
   */
  if (params->initialised == 0)
    {
#if DEBUG_EXTTRIG > 0
      /* a prefix to identify debugging messages */
      snprintf( params->id, IDLEN, "slave [node %2d]", params->rank );
#endif /*  DEBUG_EXTTRIG > 0*/

      if ( makeSpectrum)
	results->xCorrSpectra = (COMPLEX8FrequencySeries **)
	  LALCalloc( params->data->filled, sizeof(COMPLEX8FrequencySeries *) );

      params->initialised = 1;

      output->numOutput   = 0;
      output->result      = NULL;
      output->fracRemaining = 1.0;
      output->notFinished = 1;
    }

#ifdef DEBUG_EXTTRIG
      fprintf( stdout, "exttrig:LALApplySearch - %s\n", params->id );
      fflush ( stdout );
#endif /*  DEBUG_EXTTRIG*/

  mpiMsg.msg    = 0;
  mpiMsg.source = params->rank;

  /*
   * 1)  Get data from master
   */
  mpiMsg.send   = 0; /* '0' = will receive data on next MPI exchange(s) */
  LALMPISendMsg( status->statusPtr, &mpiMsg, 0, comm );
  CHECKSTATUSPTR( status );

  LALMPIRecvUINT4( status->statusPtr, &curSegment, 0, comm );
  CHECKSTATUSPTR( status );

  /* check whether we need to perform any more analysis */
  if ( curSegment >= params->data->filled )
    {
#ifdef DUMMY_CODE
      fprintf(stdout,"%s: received TERMINATE signal from master\n",params->id);
      fflush( stdout );
#endif /*  DUMMY_CODE*/

      output->numOutput     = 0;
      
#if DEBUG_EXTTRIG < 3
      /* check if output spectra are expected, if so, need to output spectra */
      if( makeSpectrum ) {
	output->numOutput   = params->result->filled;
	output->result = (outPut *)
	  LALCalloc(output->numOutput, sizeof(outPut));

	ExtTrigBuildILWD( status->statusPtr, output, params );
	CHECKSTATUSPTR( status );
      }
#else
      fprintf( stdout, "%s: finishing ApplySearch loop\n", params->id );
      fflush( stdout );
#endif /*  DEBUG_EXTTRIG < 3*/

      output->fracRemaining = 0.0;
      output->notFinished   = 0;

      DETATCHSTATUSPTR( status );
      RETURN( status );
    }
  
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: Current segment = %d\n", params->id, curSegment );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

  /*
   * 2)  Perform cross-correlation for chunk 'n' of data
   * 2a) Calculate auto-correlation statistics for chunk 'n' of data
   * 2b) Calculate cross-correlation statistic for chunk 'n' of data
   */
  /* check if output spectra are expected; if so, need to create spectra */
  onsrc = (curSegment == 0) ? 1 : 0;
  makeSpectrum = ((params->initParams->initMask & 16) && onsrc) || 
    ((params->initParams->initMask & 32) && !onsrc);
  if (makeSpectrum) {
    ccOutput.ccSpectrum = results->xCorrSpectra[results->filled] =
      (COMPLEX8FrequencySeries *) LALMalloc( sizeof(COMPLEX8FrequencySeries) );
    LALCCreateVector( status->statusPtr, &(ccOutput.ccSpectrum->data),
		      dataSeg[curSegment].filter->data->length );
    CHECKSTATUSPTR( status );

    results->filled++;
  }

  /* Auto-correlation: TimeSeries #1 */
  tmpSegment.tsChan1   = dataSeg[curSegment].tsChan1;
  tmpSegment.tsChan2   = dataSeg[curSegment].tsChan1;
  tmpSegment.filter    = dataSeg[curSegment].filter;
  tmpSegment.fftPlan   = params->fftParams;
  tmpSegment.segmentID = dataSeg[curSegment].segmentID;
  ExtTrigCrossCorrelate( status->statusPtr, &ccOutput, &tmpSegment, 0 );
  CHECKSTATUSPTR( status );
  tmpVar1 = ccOutput.ccStatistic.value;

  /* Auto-correlation: TimeSeries #2 */
  tmpSegment.tsChan1   = dataSeg[curSegment].tsChan2;
  tmpSegment.tsChan2   = dataSeg[curSegment].tsChan2;
  ExtTrigCrossCorrelate( status->statusPtr, &ccOutput, &tmpSegment, 0 );
  CHECKSTATUSPTR( status );
  tmpVar2 = ccOutput.ccStatistic.value;

  /* Cross-correlation */
  ExtTrigCrossCorrelate( status->statusPtr, &ccOutput,
			 &(dataSeg[curSegment]), makeSpectrum );
  CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 0
      fprintf( stdout,
	       "%s: tmpVar1=%8.3g;\ttmpVar2=%8.3g;c.c. statistic=%8.3g\n",
	       params->id, tmpVar1, tmpVar2, ccOutput.ccStatistic.value );
      fflush( stdout );
#endif /*  DEBUG_EXTTRIG*/


  /*
   * 3)  Send results to master
   */
  /* set up temporary storage for LALUnit information */
  LALI2CreateVector( status->statusPtr, &tmpLALUnit, 
		     (1 + 2 * LALNumUnits)          );
  CHECKSTATUSPTR( status );
  tmpLALUnit->data[0] =  ccOutput.ccStatistic.units.powerOfTen;
  memcpy( &(tmpLALUnit->data[1]), 
	  ccOutput.ccStatistic.units.unitNumerator, 
	  LALNumUnits * sizeof(INT2)      );
  memcpy( &(tmpLALUnit->data[1 + LALNumUnits]), 
	  ccOutput.ccStatistic.units.unitDenominatorMinusOne, 
	  LALNumUnits * sizeof(INT2)                );

  mpiMsg.send   = 1; /* '1' = will send data on next MPI exchange(s) */
  LALMPISendMsg( status->statusPtr, &mpiMsg, 0, comm );
  CHECKSTATUSPTR( status );

  LALMPISendUINT4( status->statusPtr, &curSegment, 0, comm );
  CHECKSTATUSPTR( status );
  
  LALMPISendREAL4( status->statusPtr, &(ccOutput.ccStatistic.value), 0, comm );
  CHECKSTATUSPTR( status );

  LALMPISendINT2Vector( status->statusPtr, tmpLALUnit, 0, comm );
  CHECKSTATUSPTR( status );

  LALMPISendREAL4( status->statusPtr, &(tmpVar1), 0, comm );
  CHECKSTATUSPTR( status );

  LALMPISendREAL4( status->statusPtr, &(tmpVar2), 0, comm );
  CHECKSTATUSPTR( status );


  /* destroy temporary storage for LALUnit */
  LALI2DestroyVector( status->statusPtr, &tmpLALUnit );
  CHECKSTATUSPTR( status );


  /*
   * 4)  FINISH
   */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

/*
 * Function ExtTrigVariance:
 * 
 * 
 * 
 * 
 */
static void
ExtTrigVariance(
		LALStatus             *const status,
		REAL4                 *const var1,
		REAL4                 *const var2,
		ETDataSegment         *const data   )
{
  REAL8                   square1=0; /* stores the expected x^2 */
  REAL8                   square2=0; /* stores the expected x^2 */
  REAL4                   mean1=0;   /* stores the mean of the time series */
  REAL4                   mean2=0;   /* stores the mean of the time series */
  REAL4TimeSeries        *ts1=data->tsChan1;     /* pointer to time series */
  REAL4TimeSeries        *ts2=data->tsChan2;     /* pointer to time series */
  INT4 i; /* iterator(s) */

  INITSTATUS( status, "ExtTrigVariance", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

  if (ts1->data->length == ts2->data->length) {
    for (i=0; i < ts1->data->length; i++) {
      mean1   += ts1->data->data[i];
      square1 += ts1->data->data[i] * ts1->data->data[i];
      mean2   += ts2->data->data[i];
      square2 += ts2->data->data[i] * ts2->data->data[i];
    }
  } else {
    ABORT( status, APPLYSEARCHH_ELGTH, APPLYSEARCHH_MSGELGTH );
  }
  mean1   /= ts1->data->length;
  square1 /= ts1->data->length;
  mean2   /= ts1->data->length;
  square2 /= ts1->data->length;
  *var1 = square1 - mean1 * mean1;
  *var2 = square2 - mean2 * mean2;

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


/*
 * Function ExtTrigCrossCorrelate:
 *
 * Called by either ApplySearchMaster or ApplySearchSlave.  This executes the
 * main search sequence-- that is, the StochasticCrossCorrelation() function.
 * It returns cross-correlation statistic via the REAL4WithUnits structure.  
 * To this end, we need the following functionality:
 * 1)  Allocate temporary memory for frequency series data
 * 2)  Zero-pad and FFT the time series data
 * 3)  Calculate the cross-correlation statistic
 * 4)  Clean up and FINISH
 */
static void
ExtTrigCrossCorrelate(
		      LALStatus                     *const status,
		      ETCrossCorrOutput             *const ccoutput,
		      const ETDataSegment           *const ccinput,
		      const BOOLEAN                  makeSpectrum   )
{
  /* Variables for LALStochasticCrossCorrelation */
  COMPLEX8FrequencySeries          fsChan1;
  COMPLEX8FrequencySeries          fsChan2;
  StochasticCrossCorrelationInput  stochInput;


  INITSTATUS( status, "ExtTrigCrossCorrelate", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

#if DEBUG_EXTTRIG > 0
  fprintf( stdout, "exttrig:ExtTrigCrossCorrelate\n" );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 0*/

  /*
   * 1)  Allocate temporary memory for frequency series data
   */
  fsChan1.data = NULL;
  LALCCreateVector( status->statusPtr, &(fsChan1.data), 
		    ccinput->filter->data->length         );
  CHECKSTATUSPTR( status );

  fsChan2.data = NULL;
  LALCCreateVector( status->statusPtr, &(fsChan2.data), 
		    ccinput->filter->data->length         );
  CHECKSTATUSPTR( status );

  /* assign frequency series parameters */
  /* fsChan1 */
  strcpy( fsChan1.name, ccinput->tsChan1->name );
  fsChan1.epoch.gpsSeconds = ccinput->tsChan1->epoch.gpsSeconds;
  fsChan1.epoch.gpsNanoSeconds = ccinput->tsChan1->epoch.gpsNanoSeconds;
  fsChan1.f0 = ccinput->filter->f0;
  fsChan1.deltaF = ccinput->filter->deltaF;
  fsChan1.sampleUnits = lalHertzUnit;
  /* fsChan2 */
  strcpy( fsChan2.name, ccinput->tsChan2->name );
  fsChan2.epoch.gpsSeconds = ccinput->tsChan2->epoch.gpsSeconds;
  fsChan2.epoch.gpsNanoSeconds = ccinput->tsChan2->epoch.gpsNanoSeconds;
  fsChan2.f0 = ccinput->filter->f0;
  fsChan2.deltaF = ccinput->filter->deltaF;
  fsChan2.sampleUnits = lalHertzUnit;

  /*
   * 2)  (Zero-pad? and) FFT the time series data
   */
  LALForwardRealFFT( status->statusPtr, fsChan1.data,
		     ccinput->tsChan1->data, ccinput->fftPlan );
  CHECKSTATUSPTR( status );

  LALForwardRealFFT( status->statusPtr, fsChan2.data,
		     ccinput->tsChan2->data, ccinput->fftPlan );
  CHECKSTATUSPTR( status );

#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "ExtTrigCrossCorrelate stats: \n" );
  fprintf( stdout, "   TSeries length = %d;\t deltaT = %g;\t f0 = %g\n",
	   ccinput->tsChan1->data->length, ccinput->tsChan1->deltaT, 
	   ccinput->tsChan1->f0 );
  fprintf( stdout, "   FSeries length = %d;\t deltaF = %g;\t f0 = %g\n",
	   fsChan1.data->length, fsChan1.deltaF, fsChan1.f0 );
  fprintf( stdout, "   Filter  length = %d;\t deltaF = %g;\t f0 = %g\n",
	   ccinput->filter->data->length, ccinput->filter->deltaF, 
	   ccinput->filter->f0 );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 2*/
#if DEBUG_EXTTRIG > 3
  /* weirdly, first call doesn't print */
  PrintFSeries( status->statusPtr, ccinput->filter, 10 );
  CHECKSTATUSPTR( status );
  PrintFSeries( status->statusPtr, ccinput->filter, 10 );
  CHECKSTATUSPTR( status );
  PrintFSeries( status->statusPtr, &fsChan1, 10 );
  CHECKSTATUSPTR( status );
  PrintFSeries( status->statusPtr, &fsChan2, 10 );
  CHECKSTATUSPTR( status );
#endif /*  DEBUG_EXTTRIG > 3*/

  /*
   * 3)  Calculate the cross-correlation statistic
   */
  stochInput.hBarTildeOne = &fsChan1;
  stochInput.hBarTildeTwo = &fsChan2;
  stochInput.optimalFilter = ccinput->filter;

  /* Calculate the cross-correlation spectrum if requested
   */
  if (makeSpectrum) {
    LALStochasticCrossCorrelationSpectrum( status->statusPtr,
					   ccoutput->ccSpectrum,
					   &stochInput, 0         );
    CHECKSTATUSPTR( status );

#if DEBUG_EXTTRIG > 2
    PrintFSeries( status->statusPtr, ccoutput->ccSpectrum, 10 );
    CHECKSTATUSPTR( status );
#endif /*  DEBUG_EXTTRIG > 2*/
  }

  LALStochasticCrossCorrelationStatistic( status->statusPtr,
					  &(ccoutput->ccStatistic),
					  &stochInput, 0          );
  CHECKSTATUSPTR( status );

  /*
   * 4)  Clean up and FINISH
   */
  LALCDestroyVector( status->statusPtr, &(fsChan2.data) );
  CHECKSTATUSPTR( status );
		     
  LALCDestroyVector( status->statusPtr, &(fsChan1.data) );
  CHECKSTATUSPTR( status );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

#if DEBUG_EXTTRIG > 0
static void
PrintFSeries(
	     LALStatus                     *const status,
	     const COMPLEX8FrequencySeries *const fseries,
	     const UINT4                    binning      )
{
  UINT4    i=0; /* iterator */
  UINT4    j=0; /* iterator */

  INITSTATUS( status, "PrintFSeries", EXTTRIGAPPLYSEARCHC );
  ATTATCHSTATUSPTR( status );

  fprintf( stdout,
	   "FSeries %s (10^%dHz):\tStart:%d.%09d\tf0:%6.2g\tdF:%6.2g\n",
	   fseries->name, fseries->sampleUnits.powerOfTen,
	   fseries->epoch.gpsSeconds, fseries->epoch.gpsNanoSeconds,
	   fseries->f0, fseries->deltaF                                );
  for (i=0; i < fseries->data->length; ) {
    fprintf( stdout, "row %4d: ", i );
    for (j=0; j < binning; j++) {
#if DEBUG_EXTTRIG < 3
      fprintf( stdout, "%10.3g ", fseries->data->data[i].re );
#else
      fprintf( stdout, "%10.3g %10.3g ",
	       fseries->data->data[i].re, fseries->data->data[i].im );
#endif /*  DEBUG_EXTTRIG < 3*/
      i++;
      if ( i == fseries->data->length ) { break; }
    }
    fprintf( stdout, "\n" );
    fflush( stdout );
  }

  DETATCHSTATUSPTR( status );
  RETURN( status );
}
#endif /*  DEBUG_EXTTRIG > 0*/

#ifdef __cplusplus
}
#endif
