/* -*- mode: c; c-basic-offset: 2; -*- */
/******** <lalVerbatim file="ExtTrigConditionDataCV"> ********
Author: Rahkola, R
$Id: ExtTrigConditionData.c,v 1.17 2003/07/03 18:30:56 jolien Exp $
********* </lalVerbatim> ********/

/* Include input/output functions if we're in the debug stage */
#ifdef DEBUG_EXTTRIG
#include <lal/PrintFTSeries.h>
#endif /*  DEBUG_EXTTRIG*/

/* standard C functions */
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <mpi.h>

/* standard LAL functions*/
#include <lal/LALStdlib.h>
#include <lal/Units.h>
#include <lal/LALConstants.h>
#include <lal/Random.h>
#include <lal/AVFactories.h>
#include <lal/VectorOps.h>
#include <lal/Window.h>
#include <lal/Date.h>
#include <lal/TimeDelay.h>
#include <lal/DetectorSite.h>
#include <lal/Comm.h>

/* LAL MPI interface */
#include <lal/Comm.h>

/* Associated headers */
#include <ExtTrig.h>
#include <ExtTrigConditionData.h>

#ifdef __cplusplus
extern "C" {
#endif

NRCSID( EXTTRIGCONDITIONDATAC, "$Id: ExtTrigConditionData.c,v 1.17 2003/07/03 18:30:56 jolien Exp $");

static void 
GetTimeSeriesIndex( 
		   LALStatus          *status, 
		   UINT4              *index,
		   LIGOTimeGPS        *gpsRef, 
		   const multiDimData *md,
		   const REAL8         delay  );

static void
AddGPSTimes(
	    LALStatus         *status,
	    LIGOTimeGPS       *output,
	    const LIGOTimeGPS *const input1,
	    const LIGOTimeGPS *const input2,
	    const BOOLEAN     subtract      );

static void
FindSegments(
	     LALStatus             *status,
	     ETSegList             *output,
	     const ETSearchParams  *const searchparams );

static void
FillFSeries(
	    LALStatus               *status,
	    const multiDimData      *const kernel,
	    const multiDimData      *const notches,
	    ETSearchParams          *const params  );

/************************* wrapperAPI function definitions  ******************/

/*
 * Function LALConditionData
 *
 * Reads the input data provided to the wrapperAPI from LDAS (usually
 * from the data-conditioning API) which is a list of sequences.
 * The code is looking for time-series data with the name "H(L)1(2):LSC-AS_Q".
 *
 */
void
LALConditionData(
    LALStatus      *status,
    LALSearchInput *inout,
    void           *searchParams,
    LALMPIParams   *mpiParams    )
{
  UINT4                 i,j;            /* iterator(s)                    */
  CHAR                  errstr[500];    /* temporary error string         */
  CHAR                  timestr[100];   /* temporary time string          */
  /* pointers for various places in the searchParams structure */
  ETSearchParams       *params    = (ETSearchParams *) searchParams;
  ETInitParams         *initParams= params->initParams;
  ETDataSegment        *dataSeg   = params->data->data;
  ETSegment            *segments  = params->segments->data;
  SearchSummaryInfo    *info      = params->info;
  /* variables for communicating between master/slave */
  MPIMessage            mpiMsg;         /* send/rec'v messages w/slave    */
  UINT4                 mpiSize = 0;    /* number of parallel nodes       */
  /* variables for assigning multiDimData */
  multiDimData         *chan1 = NULL;   /* identify chan1 in multiDimData */
  multiDimData         *chan2 = NULL;   /* identify chan2 in multiDimData */
  multiDimData         *kernel= NULL;   /* identify kernel in multiDimData*/
  multiDimData         *notch = NULL;   /* identify notch in multiDimData */
  BOOLEAN               isKernel  = ( strcmp(params->kernel,"_") &&
				      strcmp(params->kernel,"")   );
  BOOLEAN               isNotch   = ( strcmp(params->notch,"_") &&
				      strcmp(params->notch,"")   );
  /* variables for filling the Segment list */
  TwoDetsTimeAndASource *tmpConfig =    /* for calculating delays         */
    (TwoDetsTimeAndASource *) LALMalloc( sizeof(TwoDetsTimeAndASource) );
  SkyPosition          *tmpPosition = NULL; /* for calculating delays     */
  LALPlaceAndGPS       *tmpDetTime1;       /* temp pointer                */
  LALPlaceAndGPS       *tmpDetTime2;       /* temp pointer                */
  /* variables for filling the DataSegment vector */
  CHARVector           *selVector =NULL;/* fills 'selected' parameters    */
  REAL8Vector          *lagVector =NULL;/* fills 'selected' parameters    */
  UINT4                 tsIndexStart, tsIndexEnd; /* for extracted ts     */
  LIGOTimeGPS           gpsDelay;       /* for 2nd timeseries epoch start */
  LIGOTimeGPS           adjGPSStart;    /* for 2nd timeseries epoch start */
  UINT4                 kernLength=     /* expected length of freq kernel */
    initParams->segLength / 2 + 1;
  REAL4                 kernDeltaF=     /* expected freq step for kernel  */
    ((REAL4) initParams->sampleRate) / initParams->segLength;
  /* variables for picking random segments and setting random delays */
  static RandomParams  *randGen;                  /* random number        */
#if DEBUG_EXTTRIG > 0
  INT4                  seed      = 12345;        /* 0 = use current time */
#else
  INT4                  seed      = 0;            /* 0 = use current time */
#endif /*  DEBUG_EXTTRIG > 0*/
  REAL4                 randNext  = 0.0;/* the actual randomly-generated #*/
  UINT4                 randSeg   = 0;  /* a random index of segList      */
  REAL8                 randLag   = 0.0;/* a random lag for randSeg       */
  UINT4                 randLeft  = initParams->numOffSegments;
  UINT4                 initMask  = initParams->initMask;

  /* variables for filling in the FFT window */
/*    static const char    *WindowTypeNames[] = WINDOWNAMELIST; */
/*    WindowType            winName; */
/*    LALWindowParams       winParams; */


  INITSTATUS( status, "LALConditionData", EXTTRIGCONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  /*
   * check that searchParams is valid and cast it to ETSearchParams*
   */
  ASSERT( searchParams, status, CONDITIONDATAH_ENULL, 
	  CONDITIONDATAH_MSGENULL );
  params = (ETSearchParams *) searchParams;

  /*
   * some initial Node-specific code here
   */
  /* Determines whether we're running as master (0) or as a slave (>0) */
  MPI_Comm_rank( *(mpiParams->comm), &(params->rank) );
  /* get the number of nodes */
  MPI_Comm_size( *(mpiParams->comm), &(mpiSize));
  mpiSize -= 1;     /* subtract 1 to get the number of slaves */

  /* assign the SearchSummaryInfo number of nodes */
  if (mpiSize) {
    info->numNodes = mpiSize;	/* Only slaves do the search work */
  } else {
    info->numNodes = 1;		/* Now the master does the search work */
  }

#if defined (DEBUG_EXTTRIG) || defined (DUMMY_CODE)
  if( params->searchMaster )
    snprintf( params->id, IDLEN, "master[node %2d]", params->rank );
  else
    snprintf( params->id, IDLEN, "slave[node %2d]",  params->rank );
#endif /*  defined (DEBUG_EXTTRIG) || defined (DUMMY_CODE)*/
#ifdef DEBUG_EXTTRIG
  fprintf( stdout, "exttrig:LALConditionData- %s\n", params->id );
  fflush ( stdout );
#if DEBUG_EXTTRIG > 0
  fprintf( stdout, "%s: params->rank = %d,\tmpiSize = %d\n", 
	   params->id, params->rank, mpiSize                );
  fflush ( stdout );
#endif /*  DEBUG_EXTTRIG > 0*/
#endif /*  DEBUG_EXTTRIG*/
  
  /**********************
   **********************
   ***  Sanity checks on inout data
   **********************
   **********************/
  /* check that inout data exists */
  ASSERT( inout, status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );

  /* check that the number of multiDimData sequences (time series, frequency
   * series, time-frequency series, database entries, etc.) received from the 
   * datacondAPI is correct.  With two h(t) channels, an optimal filter, and
   * a sequence of notches, this could be three or four.
   * :NOTE: Currently we're expecting 2 time series, (possibly) an optimal
   * :NOTE: filter, (possibly) a sequence of notches (central frq., 1/2-width)
   */
  if ( inout->numberSequences < 2 )
    {
      ABORT( status, CONDITIONDATAH_ENSEQ, CONDITIONDATAH_MSGENSEQ );
    }

  /* check that the multidimData sequence data exists */
  if ( !(inout->sequences) )
    {
      ABORT( status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );
    }


  /**********************
   **********************
   ***  Memory allocation and initialization
   ***  (stuff which can be taken care of _before_ reading multiDimData)
   **********************
   **********************/
  /* REAL4Sequence for FFT window
   */
/*    for (winName = NumberWindowTypes; winName != Rectangular; winName--) { */
/*      if (!strcmp(initParams->window,WindowTypeNames[winName])) break; */
/*    } */
/*    winParams.type   = winName; */
/*    winParams.length = params->fftParams->length; */
/*  #if DEBUG_EXTTRIG > 1 */
/*    fprintf( stdout, "%s: Window type = %s\tlength = %d\n", params->id, */
/*  	   WindowTypeNames[winName], winParams.length ); */
/*    fflush( stdout ); */
/*  #endif /*  DEBUG_EXTTRIG > 1 */
/*    LALWindow( status->statusPtr, params->fftParams->window, &winParams ); */
/*    CHECKSTATUSPTR( status ); */

  /* CHARVector for segment selection
   */
  LALCHARCreateVector( status->statusPtr, &selVector, 
		       params->segments->length );
  CHECKSTATUSPTR( status );
  for (i = 0; i < selVector->length; i++)
    selVector->data[i] = 0; /* make sure everything's cleared out first */

  /* REAL8Vector for lag specification
   */
  LALDCreateVector( status->statusPtr, &lagVector, params->segments->length );
  CHECKSTATUSPTR( status );
  for (i = 0; i < lagVector->length; i++)
    lagVector->data[i] = 0; /* make sure everything's cleared out first */

  /* RandomParams RandNext for generating random numbers
   */
  LALCreateRandomParams( status->statusPtr, &randGen, seed );
  CHECKSTATUSPTR( status );

  /* tmpConfig (needs two LALPlaceAndGPS structures)
   */
  tmpDetTime1 = tmpConfig->p_det_and_time1 = (LALPlaceAndGPS *) 
    LALMalloc( sizeof(LALPlaceAndGPS) );
  tmpDetTime2 = tmpConfig->p_det_and_time2 = (LALPlaceAndGPS *) 
    LALMalloc( sizeof(LALPlaceAndGPS) );
  /* tentatively point to NULL, but we'll change it after FindSegments() */
  tmpConfig->p_source = NULL;

  /* first LALPlaceAndGPS structure
   */
  tmpDetTime1->p_detector = (LALDetector *) LALMalloc( sizeof(LALDetector) );
  /* temporarily point to NULL, but we'll change it in FindSegments() */
  tmpDetTime1->p_gps = NULL;

  /* second LALPlaceAndGPS structure
   */
  tmpDetTime2->p_detector = (LALDetector *) LALMalloc( sizeof(LALDetector) );
  /* permanently point to NULL, since it won't be used while calculating 
     time delay */
  tmpDetTime2->p_gps = NULL;

  /* tmpPosition (needs two REAL8's and an enumeration)
   */
  tmpPosition = (SkyPosition *) LALMalloc( sizeof(SkyPosition) );
  tmpPosition->system = initParams->sourceDir.system;


  /**********************
   **********************
   *** identify the multiDimData sequences by their names (specified in the Tcl
   *** script), and check that the data has the correct type, dimensions, and
   *** length
   **********************
   **********************/
  for ( i=0; i < inout->numberSequences; ++i )
    {
      switch (inout->sequences[i].space) {
      /*****************
       *****************
       ***  Time domain multiDimData (channel 1, channel 2)
       *****************
       *****************/
      case timeD:
	/* Matching channel 1
	 */
	if ( strstr (inout->sequences[i].comment, params->chan1) ) {
	  /* assign an index number for reference later on */
	  chan1 = &(inout->sequences[i]);
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: Assigning chan1 to multiDimData %d: %s\n", 
		   params->id, i, chan1->comment );
	  fprintf( stdout,
		   "%s: chan1 multiDimData stats: numDim: %d\n",
		   params->id, chan1->numberDimensions );
	  fflush (stdout);
#endif /*  DEBUG_EXTTRIG > 1*/

	  /* assign the SearchSummaryInfo times corresponding to channel 1 */
	  info->inputStartTime.gpsSeconds = chan1->range.dTime.startSec;
	  info->inputStartTime.gpsNanoSeconds = chan1->range.dTime.startNan;
	  info->inputEndTime.gpsSeconds = chan1->range.dTime.stopSec;
	  info->inputEndTime.gpsNanoSeconds = chan1->range.dTime.stopNan;

	/* Matching channel 2
	 */
	} else if ( strstr (inout->sequences[i].comment, params->chan2) ) {
	  /* assign an index number for reference later on */
	  chan2 = &(inout->sequences[i]);
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: Assigning chan2 to multiDimData %d: %s\n",
		   params->id, i, chan2->comment );
	  fprintf( stdout,
		   "%s: chan2 multiDimData stats: numDim: %d\n",
		   params->id, chan2->numberDimensions );
	  fflush (stdout);
#endif
	} else break;

	/* check dimensions */
	if ( inout->sequences[i].numberDimensions != 1 ) {
	  ABORT (status, CONDITIONDATAH_EDIMS, CONDITIONDATAH_MSGEDIMS );
	}

	/* check type of data (need REAL4) */
	if ( inout->sequences[i].type != real_4 ) {
	  ABORT( status, CONDITIONDATAH_ETYPE, CONDITIONDATAH_MSGETYPE );
	}
	/* :NOTE: Don't actually check length now, but should in the future*/
	break;


      /*****************
       *****************
       ***  Frequency domain multiDimData (filter kernel, 
       *****************
       *****************/
      case freqD:
	/* Matching filter kernel
	 */
	if (isKernel && strstr(inout->sequences[i].comment, params->kernel)) {
	  /* assign an index number for reference later on */
	  kernel = &(inout->sequences[i]);
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: Assigning kernel to multiDimData %d: %s\n",
		   params->id, i, kernel->comment );
	  fprintf( stdout,
		   "%s: kernel multiDimData stats: numDim: %d, dim[0]: %d\n",
		   params->id, kernel->numberDimensions,
		   kernel->dimensions[0]                                     );
	  fflush (stdout);
#endif /*  DEBUG_EXTTRIG > 1*/
	} else break;

	/* check dimensions */
	if ( inout->sequences[i].numberDimensions != 1 ) {
	  ABORT (status, CONDITIONDATAH_EDIMS, CONDITIONDATAH_MSGEDIMS );
	}

	/* check type of data (need COMPLEX8, but will accept COMPLEX16) */
	if ( (inout->sequences[i].type != complex_8) &&
	     (inout->sequences[i].type != complex_16) ) {
	  ABORT( status, CONDITIONDATAH_ETYPE, CONDITIONDATAH_MSGETYPE );
	}

	/* check length (must match kernLength!) */
	if ( inout->sequences[i].dimensions[0] != kernLength ) {
	  ABORT( status, CONDITIONDATAH_ELGTH, CONDITIONDATAH_MSGELGTH );
	}

	/* check frequency spacing of data (must match kernDeltaF!) */
	if ( inout->sequences[i].range.dFreq.freqStepSize != kernDeltaF ) {
	  ABORT( status, CONDITIONDATAH_ESPC, CONDITIONDATAH_MSGESPC );
	}

	break;


      /******************
       ******************
       ***  Raw sequence domain multiDimData (notch frequencies, 
       ******************
       ******************/
      case noneD:
	/* Matching frequency notch info
	 */
	if (isNotch && strstr(inout->sequences[i].comment, params->notch)) {
	  /* assign an index number for reference later on */
	  notch = &(inout->sequences[i]);
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: Assigning notch data to multiDimData %d: %s\n",
		   params->id, i, notch->comment );
	  fprintf( stdout,
		   "%s: notch multiDimData stats: numDim: %d, dim[0]: %d\n",
		   params->id, notch->numberDimensions, notch->dimensions[0] );
	  fflush (stdout);
#endif /*  DEBUG_EXTTRIG > 1*/
	} else break;

	/* check dimensions */
	if ( inout->sequences[i].numberDimensions != 1 ) {
	  ABORT (status, CONDITIONDATAH_EDIMS, CONDITIONDATAH_MSGEDIMS );
	}

	/* check type of data (need REAL4 or REAL8) */
	if ( (inout->sequences[i].type != real_4) &&
	     (inout->sequences[i].type != real_8)   ) {
	  ABORT( status, CONDITIONDATAH_ETYPE, CONDITIONDATAH_MSGETYPE );
	}

	break;


      /* :TODO: Handle data quality multiDimData for both IFOs */
      default:
	break;
      } /*   switch (inout->sequences[i].space) ...*/
    } /*   for ( i=0; i < inout->numberSequences; ++i ) ...*/


  /**********************
   **********************
   ***  check that all necessary multiDimData has been assigned
   ***  :TODO: Need to include data quality channel checks here
   **********************
   **********************/
  if ( (chan1) && (chan2) && (kernel || !isKernel) && (notch || !isNotch) ) ;
  else {
      ABORT( status, CONDITIONDATAH_EFIND, CONDITIONDATAH_MSGEFIND );
  }


  /**********************
   **********************
   *** (Allocate and) Assign some multiDimData-dependent variables
   **********************
   **********************/
  /* params->tsChan1 and params->tsChan2
   */
  strncpy( params->tsChan1->name, params->chan1, dbNameLimit );
  strncpy( params->tsChan2->name, params->chan2, dbNameLimit );
  params->tsChan1->epoch.gpsSeconds     = chan1->range.dTime.startSec;
  params->tsChan1->epoch.gpsNanoSeconds = chan1->range.dTime.startNan;
  params->tsChan2->epoch.gpsSeconds     = chan2->range.dTime.startSec;
  params->tsChan2->epoch.gpsNanoSeconds = chan2->range.dTime.startNan;
  params->tsChan1->deltaT               = chan1->range.dTime.timeStepSize;
  params->tsChan2->deltaT               = chan2->range.dTime.timeStepSize;
  params->tsChan1->f0                   = chan1->range.dTime.baseFreq;
  params->tsChan2->f0                   = chan2->range.dTime.baseFreq;
  params->tsChan1->sampleUnits          = lalADCCountUnit;
  params->tsChan2->sampleUnits          = lalADCCountUnit;
  LALSCreateVector( status->statusPtr, &params->tsChan1->data,
		       chan1->dimensions[0] );
  CHECKSTATUSPTR( status );
  LALSCreateVector( status->statusPtr, &params->tsChan2->data,
		       chan2->dimensions[0] );
  CHECKSTATUSPTR( status );
  memcpy( params->tsChan1->data->data, chan1->data.real4,
	  params->tsChan1->data->length * sizeof(REAL4) );
  memcpy( params->tsChan2->data->data, chan2->data.real4,
	  params->tsChan2->data->length * sizeof(REAL4) );


  /* params->fsKernel and params->fsNotch
   */
  FillFSeries( status->statusPtr, kernel, notch, params );
  CHECKSTATUSPTR( status );

  /**********************
   ***  tmpConfig detectors -- This is complicated by the fact that I'd like
   ***  to use the cached detector configurations.  I break up this
   ***  assignment into three cases:
   ***   1) If I'm in debug mode (DEBUG_EXTTRIG > 0), I assign ifo1 to LHO
   ***      and ifo2 to LLO
   ***   2) If I'm not in debug mode, I check the first letter of the channel
   ***      names (corresponding to the comment field in multiDimData).
   ***      Currently I only recognize Livingston and Hanford detectors.
   ***   3) If I can't recognize the detector by the first letter, I use the
   ***      multiDimData detGeom structure.
   *** :TO-DO: Actually use the multiDimData structure when I can't determine
   *** :TO-DO: the detector by its channel name.
   **********************/
#if DEBUG_EXTTRIG > 0
  /* Use LHO and LLO as det1 and det2, respectively */
  *(tmpDetTime1->p_detector) = lalCachedDetectors[LALDetectorIndexLHODIFF];
  *(tmpDetTime2->p_detector) = lalCachedDetectors[LALDetectorIndexLLODIFF];
#else
  /* choose the two detectors based on channel names */
  switch(params->chan1[0]) {
  case 'H':
    *(tmpDetTime1->p_detector) = lalCachedDetectors[LALDetectorIndexLHODIFF];
    break;
  case 'L':
    *(tmpDetTime1->p_detector) = lalCachedDetectors[LALDetectorIndexLLODIFF];
    break;
  default:
    /* :TODO: Assign detector based on detGeom structure in ILWD file */
    sprintf( errstr, CONDITIONDATAH_MSGEDET);
    strcat( errstr, params->chan1 );
    ABORT( status, CONDITIONDATAH_EDET, errstr );
  }
  switch(params->chan2[0]) {
  case 'H':
    *(tmpDetTime2->p_detector) = lalCachedDetectors[LALDetectorIndexLHODIFF];
    break;
  case 'L':
    *(tmpDetTime2->p_detector) = lalCachedDetectors[LALDetectorIndexLLODIFF];
    break;
  default:
    /* :TODO: Assign detector based on detGeom structure in ILWD file */
    sprintf( errstr, CONDITIONDATAH_MSGEDET);
    strcat( errstr, params->chan2 );
    ABORT( status, CONDITIONDATAH_EDET, errstr );
  }
#endif /*  DEBUG_EXTTRIG > 0*/
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: Det1: %s,\tDet2: %s\n", params->id,
	   tmpDetTime1->p_detector->frDetector.name,
	   tmpDetTime2->p_detector->frDetector.name );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/


  /* Segment list (timestamps only, for now)
   */
  FindSegments( status->statusPtr, params->segments, params );
  CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 1
  for( i = 0; i < params->segments->length; i++ )
    {
      fprintf( stdout, "%s: seg%2d-- startGPS: %d.%09d;\tendGPS: %d.%09d;\tdelay: %6.5f\n",
	       params->id, i, params->segments->data[i].startGPS.gpsSeconds,
	       params->segments->data[i].startGPS.gpsNanoSeconds,
	       params->segments->data[i].endGPS.gpsSeconds,
	       params->segments->data[i].endGPS.gpsNanoSeconds,
	       params->segments->data[i].delay                             );
      fflush( stdout );
    }
#endif /*  DEBUG_EXTTRIG > 1*/


  /* SearchSummaryInfo (start and end timestamps)
   */
  info->outputStartTime.gpsSeconds =
    params->segments->data[0].startGPS.gpsSeconds;
  info->outputStartTime.gpsNanoSeconds =
    params->segments->data[0].startGPS.gpsNanoSeconds;
  info->outputEndTime.gpsSeconds =
    params->segments->data[params->segments->length - 1].endGPS.gpsSeconds;
  info->outputEndTime.gpsNanoSeconds =
    params->segments->data[params->segments->length - 1].endGPS.gpsNanoSeconds;


  /**********************
   **********************
   *** (Master only): Pick the segments which will be used and pass to slaves
   *** (Slaves only): Receive the 'selected' info for the segments
   **********************
   **********************/
  if( params->searchMaster )
    {
      selVector->data[0] = 1; /* first segment is always selected */
      /* set the lag for on-src segment */
      /* test case:  2^3 -- fix lag using sourceDir.longitude (sourceRA) */
      if (initParams->initMask & 8) {
	lagVector->data[0] = initParams->sourceDir.longitude;
      /* test case: !2^3 -- fix lag using SkyPosition */
      } else {
	tmpConfig->p_source = (SkyPosition *) &(initParams->sourceDir);
	tmpDetTime1->p_gps = &(segments[0].startGPS);
	tmpDetTime2->p_gps = &(segments[0].startGPS);
	LALTimeDelay( status->statusPtr, &(lagVector->data[0]), tmpConfig );
	CHECKSTATUSPTR( status );
      }

      /* keep picking random numbers until the desired number is obtained,
       * or there are no more segments to pick from
       */
      if (randLeft > params->segments->filled) 
	randLeft = params->segments->filled;
#if DEBUG_EXTTRIG > 1
      fprintf( stdout, "%s: # off-segs:%d,\t# filled:%d,\trandLeft = %d\n",
	       params->id, initParams->numOffSegments,
	       params->segments->filled, randLeft  );
      fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
      while (randLeft > 0)
	{
	  /*  First we'll use randNext to identify a segment [1,NSeg] to tag...*/
	  LALUniformDeviate( status->statusPtr, &randNext, randGen );
	  CHECKSTATUSPTR( status );
	  randSeg = (INT4)(randNext * (params->segments->length - 1)) + 1;
#if DEBUG_EXTTRIG > 2
	  fprintf( stdout, "%s: random segment selection: %3d\n", 
		   params->id, randSeg );
	  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 2*/
	  /* check if we've selected this segment already */
	  if( (randSeg > selVector->length) || selVector->data[randSeg] ) 
	    continue;
          /* test case:  2^2 &  2^3 -- fixed lag, using sourceDir.latitude */
	  if (initMask & 4) {
	    if (initMask & 8) randLag = initParams->sourceDir.latitude;
	  /* test case:  2^2 & !2^3 -- fixed lag, using 0.0 */
	    else  randLag = (REAL8)0.0;
	  /* test case: !2^2 &  2^3 -- random lag, uniformly dist. in time */
	  } else {
	    tmpConfig->p_source = tmpPosition;
	    if (initMask & 8) {
	      LALUniformDeviate( status->statusPtr, &randNext, randGen );
	      CHECKSTATUSPTR( status );
	      /* split over +/- [segMinDelay,segMaxDelay] */
	      randLag = ( (2 * randNext - 1.0) *
			  (initParams->segMaxDelay -
			   initParams->segMinDelay ) /
			  ((REAL8) initParams->sampleRate) );
	      if (randLag < 0.0) randLag -=
		    ((REAL8) initParams->segMinDelay) / initParams->sampleRate;
	      else randLag +=
		    ((REAL8) initParams->segMinDelay) / initParams->sampleRate;
	   /* test case: !2^2 & !2^3 -- random lag, uniformly dist. in space */
	    } else {
	      /* set random source declination */
	      LALUniformDeviate( status->statusPtr, &randNext, randGen );
	      CHECKSTATUSPTR( status );
	      randNext = (LAL_PI * randNext) - LAL_PI_2;
	      tmpPosition->latitude = randNext;

	      /* set random source right ascension */
	      LALUniformDeviate( status->statusPtr, &randNext, randGen );
	      CHECKSTATUSPTR( status );
	      randNext *= LAL_2_PI;
	      tmpPosition->longitude = randNext;

	      LALTimeDelay( status->statusPtr, &randLag, tmpConfig );
	      CHECKSTATUSPTR( status );
	    }
	  }
	  
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: selecting segment %2d w/delay %7.5f,\trandLeft = %d\n", 
		   params->id, randSeg, randLag, randLeft );
	  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
	  selVector->data[randSeg] = 1;
	  lagVector->data[randSeg] = randLag;
	  randLeft--;
	}

      /* now wait for each slave to check in and send them the boolean vector
       */
      while (mpiSize > 0)
	{
	  LALMPIRecvMsg( status->statusPtr, &mpiMsg, *(mpiParams->comm) );
	  CHECKSTATUSPTR( status );
	  if( !mpiMsg.send )
	    {
	      LALMPISendCHARVector( status->statusPtr, selVector, 
				    mpiMsg.source, *(mpiParams->comm) );
	      CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 1
	      fprintf( stdout, 
	               "%s: sent selVector to slave [node %d] \n",
		       params->id, mpiMsg.source );
	      fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
	      LALMPISendREAL8Vector( status->statusPtr, lagVector, 
				     mpiMsg.source, *(mpiParams->comm) );
	      CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 1
	      fprintf( stdout, 
	               "%s: sent lagVector to slave [node %d] \n",
		       params->id, mpiMsg.source );
	      fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
	    }
	  else /* shouldn't happen, slaves should be begging for data! */
	    {
	      ABORT( status, CONDITIONDATAH_EMPI, CONDITIONDATAH_MSGEMPI );
	    }
	  mpiSize--;
	}
    }
  else /* must be a slave; just get the two segment vectors */
    {
      /* setup the mpiMsg structure 
       */
      mpiMsg.msg    = 0;
      mpiMsg.source = params->rank;
      mpiMsg.send   = 0; /* '0' = will receive data on next MPI exchange(s) */

      /* connect to master and retrieve the vectors
       */
      LALMPISendMsg( status->statusPtr, &mpiMsg, 0, *(mpiParams->comm) );
      CHECKSTATUSPTR( status );
      LALMPIRecvCHARVector( status->statusPtr,selVector,0,*(mpiParams->comm) );
      CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 1
      fprintf( stdout, "%s: rec'vd selVector from master\n",
	       params->id );
      fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
      LALMPIRecvREAL8Vector( status->statusPtr,lagVector,0,*(mpiParams->comm));
      CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 1
      fprintf( stdout, "%s: rec'vd lagVector from master\n",
	       params->id );
      fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
    }

  /*
   * fill in the 'selected' parameter for each segment (both master and slave)
   */
  for( i=0; i < selVector->length; i++) {/* selVector.size = segments.size */
    segments[i].selected = selVector->data[i];
    segments[i].delay    = lagVector->data[i];
  }
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: Selected segments:", params->id );
  for( i=0; i < selVector->length; i++)
    fprintf( stdout, "\t%-8d", selVector->data[i] );
  fprintf( stdout, "\n" );
  fprintf( stdout, "%s: Selected lags:    ", params->id );
  for( i=0; i < lagVector->length; i++)
    fprintf( stdout, "\t%+8.2e", lagVector->data[i] );
  fprintf( stdout, "\n" );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/


  /**********************
   **********************
   *** fill ETDataSegments with h(t) multiDimData and optimal filter data
   *** using ETSegList information
   *** :NOTE: We're assuming that the data is REAL4. If not, we should
   *** :NOTE:  have aborted earlier.
   *** we're going to use two iterators here.  'j' refers to the ETDataSegment,
   *** while 'i' refers to the ETSegment we're running over.  We'll pick out
   *** time series for any ETSegment[i] which was selected (above), and stick
   *** it into ETDataSegment[j], then increment j.
   **********************
   **********************/
  params->data->segsToBeAnalyzed = params->data->filled = j = 0;
  for (i = 0; i < params->segments->length; ++i)
    {
      if ( segments[i].selected ) 
	{
	  /*
	   * segmentID
	   */
	  dataSeg[j].segmentID = i;


	  /**************
	   **************
	   *** tsChan1
	   **************
	   **************/
	  GetTimeSeriesIndex( status->statusPtr, &(tsIndexStart), 
			      &(segments[i].startGPS), chan1 , 0.0 );
	  CHECKSTATUSPTR( status );
	  GetTimeSeriesIndex( status->statusPtr, &(tsIndexEnd), 
			      &(segments[i].endGPS), chan2, 0.0 );
	  CHECKSTATUSPTR( status );
	  if ( tsIndexEnd !=
	       tsIndexStart + dataSeg[j].tsChan1->data->length ) {
	    ABORT( status, CONDITIONDATAH_ETIME, CONDITIONDATAH_MSGETIME );
	  }
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, 
     "%s: seg %d; tsChan1 first = %d\ttsChan1 last = %d\t# points = %d,%d\n",
		   params->id, j, tsIndexStart, tsIndexEnd - 1, 
		   (tsIndexEnd - tsIndexStart), initParams->segLength );
	  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

	  /*** :NOTE: We used to have a memcpy statement here, but now we are
	   *** :NOTE: only pointing the data segments to the appropriate place
	   *** :NOTE: in the full TimeSeries located in ETSearchParams.
	   ***/
	  dataSeg[j].tsChan1->data->data =
	    &(params->tsChan1->data->data[tsIndexStart]);
	  strncpy( dataSeg[j].tsChan1->name, params->chan1, dbNameLimit );
	  dataSeg[j].tsChan1->epoch = segments[i].startGPS;/*LIGOTimeGPS*/
	  dataSeg[j].tsChan1->deltaT =                  /* REAL8       */
	    chan1->range.dTime.timeStepSize;
	  dataSeg[j].tsChan1->f0 =                      /* REAL8       */
	    chan1->range.dTime.baseFreq;
	  /*
	   * :TODO: Need to read units from the multiDimData
	   */
	  dataSeg[j].tsChan1->sampleUnits = lalADCCountUnit;/* LALUnit */

	  
	  /**************
	   **************
	   *** tsChan2
	   **************
	   **************/
	  LALFloatToGPS( status->statusPtr,&(gpsDelay),&(segments[i].delay) );
	  CHECKSTATUSPTR( status );
	  AddGPSTimes( status->statusPtr, &(adjGPSStart),
		       &(segments[i].startGPS), &(gpsDelay),0);
	  CHECKSTATUSPTR( status );
	  GetTimeSeriesIndex( status->statusPtr, &(tsIndexStart), 
			      &(segments[i].startGPS), chan2,
			      segments[i].delay                  );
	  CHECKSTATUSPTR( status );
	  GetTimeSeriesIndex( status->statusPtr, &(tsIndexEnd), 
			      &(segments[i].endGPS), chan2, 
			      segments[i].delay               );
	  CHECKSTATUSPTR( status );
	  if ( tsIndexEnd !=
	       tsIndexStart + dataSeg[j].tsChan2->data->length ) {
	    ABORT( status, CONDITIONDATAH_ETIME, CONDITIONDATAH_MSGETIME );
	  }
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, 
      "%s: seg %d; tsChan2 first = %d\ttsChan2 last = %d\t# points = %d,%d\n",
		   params->id, j, tsIndexStart, tsIndexEnd - 1,
		   (tsIndexEnd - tsIndexStart), initParams->segLength );
	  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

	  /*** :NOTE: We used to have a memcpy statement here, but now we are
	   *** :NOTE: only pointing the data segments to the appropriate place
	   *** :NOTE: in the full TimeSeries located in ETSearchParams.
	   ***/
	  dataSeg[j].tsChan2->data->data =
	    &(params->tsChan2->data->data[tsIndexStart]);
	  strncpy( dataSeg[j].tsChan2->name, params->chan2, dbNameLimit );
	  dataSeg[j].tsChan2->epoch = adjGPSStart;       /* LIGOTimeGPS */
	  dataSeg[j].tsChan2->deltaT =                   /* REAL8       */
	    chan2->range.dTime.timeStepSize;
	  dataSeg[j].tsChan2->f0 =                       /* REAL8       */
	    chan2->range.dTime.baseFreq;
	  /*
	   * :TODO: Need to read units from the multiDimData
	   */
	  dataSeg[j].tsChan2->sampleUnits = lalADCCountUnit; /* LALUnit */


	  /**************
	   **************
	   *** filter kernel
	   **************
	   **************/
	  dataSeg[j].filter = params->fsKernel;
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: seg %d; kernel length = %d\tf0 = %6.2g\t deltaF = %6.2g\n",
		   params->id, j, dataSeg[j].filter->data->length ,
		   dataSeg[j].filter->f0, dataSeg[j].filter->deltaF );
	  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/


	  /**************
	   **************
	   *** notch
	   **************
	   **************/
	  dataSeg[j].notch = params->fsNotch;
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: seg %d; notch  length = %d\tf0 = %6.2g\t deltaF = %6.2g\n",
		   params->id, j, dataSeg[j].notch->data->length ,
		   dataSeg[j].notch->f0, dataSeg[j].notch->deltaF );
	  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/


	  /**************
	   **************
	   *** fftPlan
	   **************
	   **************/
	  dataSeg[j].fftPlan = params->fftParams;


	  /*
	   * :TODO: Need to copy over data quality information
	   */

	  
	  /* increment j,data->filled, and data->segsToBeAnalyzed
	   *when successful 
	   */
	  j++;
	  params->data->filled++;
	  params->data->segsToBeAnalyzed++;
	} /* if ( segments[i].selected )...              */
    } /* for ( i = 0; i < params->segments->length; i++ )... */
      
  /* free temporary allocated memory */
  LALFree( tmpPosition );
  LALFree( tmpDetTime2->p_detector );
  LALFree( tmpDetTime1->p_detector );
  LALFree( tmpDetTime2 );
  LALFree( tmpDetTime1 );
  LALDestroyRandomParams( status->statusPtr, &randGen );
  CHECKSTATUSPTR( status );
  LALDDestroyVector( status->statusPtr, &lagVector );
  CHECKSTATUSPTR( status );
  LALCHARDestroyVector( status->statusPtr, &selVector );
  CHECKSTATUSPTR( status );
  LALFree( tmpConfig );
  tmpConfig = NULL;

  /* exit normally */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

static void 
GetTimeSeriesIndex( 
		   LALStatus          *status, 
		   UINT4              *index,
		   LIGOTimeGPS        *gpsRef, 
		   const multiDimData *md,
		   const REAL8         delay  )
{
  REAL8         tsStop;
  REAL8         tsStart;
  REAL8         gpsTime;
  LIGOTimeGPS   multiDimEpoch;
  CHAR          errString[500];
  CHAR          timeString[100];

  INITSTATUS( status, "GetTimeSeriesIndex", EXTTRIGCONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  LALGPStoFloat( status->statusPtr, &(gpsTime), gpsRef );
  CHECKSTATUSPTR( status );
  gpsTime += delay;

  multiDimEpoch.gpsSeconds     = md->range.dTime.startSec;
  multiDimEpoch.gpsNanoSeconds = md->range.dTime.startNan;
  
  LALGPStoFloat( status->statusPtr, &(tsStart), &(multiDimEpoch) );
  CHECKSTATUSPTR( status );

/*    tsStop = tsStart + md->range.dTime.timeStepSize * (md->dimensions[0] - 1); */
  multiDimEpoch.gpsSeconds     = md->range.dTime.stopSec;
  multiDimEpoch.gpsNanoSeconds = md->range.dTime.stopNan;
  
  LALGPStoFloat( status->statusPtr, &(tsStop), &(multiDimEpoch) );
  CHECKSTATUSPTR( status );

#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "GetTimeSeriesIndex: gpsTime = %20.9f\n", gpsTime );
  fprintf( stdout, "                    tsStart = %20.9f\n", tsStart );
  fprintf( stdout, "                     tsStop = %20.9f\n", tsStop  );
  fprintf( stdout, "               timeStepSize = %20.9f\n", 
	   md->range.dTime.timeStepSize );
  fprintf( stdout, "(gpsTime - tsStart) / timeStepSize = %d\n",
	   (UINT4) ((gpsTime - tsStart) / md->range.dTime.timeStepSize) );
  fflush( stdout );
#endif  /*  DEBUG_EXTTRIG > 2*/

  if ( (tsStart > gpsTime) || (tsStop < gpsTime) )
    {
      sprintf( timeString, " gpsRef: %15.5f, tsStart: %15.5f, tsStop: %15.5f",
	       gpsTime, tsStart, tsStop );
      sprintf( errString, CONDITIONDATAH_MSGETIME );
      strcat( errString, timeString );
      ABORT( status, CONDITIONDATAH_ETIME, errString );
    }

  *index = (UINT4) ( (gpsTime - tsStart) / md->range.dTime.timeStepSize );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

static void
AddGPSTimes(
	    LALStatus         *status,
	    LIGOTimeGPS       *output,
	    const LIGOTimeGPS *const input1,
	    const LIGOTimeGPS *const input2,
	    const BOOLEAN     subtract     ) {
  INT4 tmpGPSSec  = input1->gpsSeconds + input2->gpsSeconds;
  INT4 tmpGPSNSec = input1->gpsNanoSeconds + input2->gpsNanoSeconds;
  INT4 tmpQuot = 0; /* "quotient" for carry-over from nanoseconds to seconds */

  INITSTATUS( status, "AddGPSTimes", EXTTRIGCONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  /* modification for subtraction */
  if (subtract) {
    tmpGPSSec -= 2 * input2->gpsSeconds;
    tmpGPSNSec -= 2 * input2->gpsNanoSeconds;
  }

  /* carry the 1 */
  if( input1->gpsSeconds < 0 ) tmpGPSNSec -= (2 * input1->gpsNanoSeconds);
  if( input2->gpsSeconds < 0 ) tmpGPSNSec -= (2 * input2->gpsNanoSeconds);
  if( abs(tmpGPSNSec) >= 1000000000 ) tmpQuot = tmpGPSNSec / 1000000000 ;
  else if( tmpGPSNSec < 0 && tmpGPSSec > 0 ) tmpQuot = -1;
  else if( tmpGPSNSec > 0 && tmpGPSSec < 0 ) tmpQuot = 1;

  output->gpsSeconds     = tmpGPSSec + tmpQuot;
  output->gpsNanoSeconds = abs(tmpGPSNSec - (tmpQuot * 1000000000));
  
#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "AddGPSTimes: %9d.%09d + %9d.%09d = %9d.%09d\n", 
	   input1->gpsSeconds, input1->gpsNanoSeconds,
	   input2->gpsSeconds, input2->gpsNanoSeconds,
	   output->gpsSeconds, output->gpsNanoSeconds );
  fflush( stdout );
#endif  /*  DEBUG_EXTTRIG > 2*/

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

static void
FindSegments(
	     LALStatus             *status,
	     ETSegList             *output,
	     const ETSearchParams  *const searchparams ) {
  /* pointers for various places in the searchParams structure */
  ETInitParams         *params= searchparams->initParams;
  /* variables for filling the Segments list */
  UINT4                 i;              /* iterator(s)                    */
  LIGOTimeGPS           tmpGPS;         /* create a GPS time for a delay  */
  ETSegment            *segments = output->data;

  INITSTATUS( status, "FindSegments", EXTTRIGCONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  /*
   *
   * Accumulate a list of possible segments ('on'-source == index 0)
   *
   */

  /* initial ('on'-source) segment
   */
  /* set the start GPS time of the segment */
  tmpGPS.gpsSeconds     = (INT4) ((REAL4) params->onTimeShift /
				  params->sampleRate             );
  tmpGPS.gpsNanoSeconds = (INT4) ( (params->onTimeShift % params->sampleRate) /
				   ((REAL4) params->sampleRate) * 1.0E9      );
  AddGPSTimes( status->statusPtr, &(segments[0].startGPS), /* subtraction */
		      &(params->grbStart), &tmpGPS, 1        );
  CHECKSTATUSPTR( status );
  /* set the end GPS time of the segment */
  tmpGPS.gpsSeconds     = (INT4) (params->segLength / params->sampleRate);
  tmpGPS.gpsNanoSeconds = (INT4) ( (params->segLength % params->sampleRate) /
				   ((REAL4) params->sampleRate) * 1.0E9     );
  AddGPSTimes( status->statusPtr, &(segments[0].endGPS),
	       &(segments[0].startGPS), &tmpGPS, 0      );
  CHECKSTATUSPTR( status );
  /* 'on'-source segments are always selected! */
  segments[0].selected = 1;
#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "%s -- FindSegments...\n", searchparams->id );
  fprintf( stdout, 
	   "seg%2d-- startGPS: %d.%09d;\tendGPS: %d.%09d;\tdelay: %6.5f\n", 0,
	   segments[0].startGPS.gpsSeconds,
	   segments[0].startGPS.gpsNanoSeconds,
	   segments[0].endGPS.gpsSeconds,
	   segments[0].endGPS.gpsNanoSeconds,
	   segments[0].delay                 );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 2*/
  output->filled = 1; /* keep track of how many segments are filled */

  /* rest of the ('off'-source) segments
   */
  tmpGPS.gpsSeconds     = (INT4) (params->offTimeShiftLow /
				  params->sampleRate); /* initialize tmpTime */
  tmpGPS.gpsNanoSeconds = (INT4) ( (params->offTimeShiftLow % 
				    params->sampleRate) /
				   ((REAL4) params->sampleRate) *  1.0E9 );
  for( i = 1; i < output->length; i++ )
    {
      /* set the start GPS time of the segment */
      if (i == 1) { /* only reference grbStart for first off-source segment */
	AddGPSTimes( status->statusPtr, &(segments[i].startGPS),
		     &(params->grbStart), &tmpGPS, 0            );
	CHECKSTATUSPTR( status );
      } else {             /* otherwise use the previous segment's end time */
	AddGPSTimes( status->statusPtr, &(segments[i].startGPS),
		     &(segments[i-1].endGPS), &tmpGPS, 0        );
	CHECKSTATUSPTR( status );
      }
      /* set the end GPS time of the segment */
      tmpGPS.gpsSeconds     = (INT4) (params->segLength / params->sampleRate);
      tmpGPS.gpsNanoSeconds = (INT4) ( (params->segLength %
					params->sampleRate) /
				       ((REAL4) params->sampleRate) * 1.0E9 );
      AddGPSTimes( status->statusPtr, &(segments[i].endGPS),
		   &(segments[i].startGPS), &tmpGPS, 0      );
      CHECKSTATUSPTR( status );

#if DEBUG_EXTTRIG > 2
      fprintf( stdout,
	       "seg%2d-- startGPS: %d.%09d;\tendGPS: %d.%09d;\tdelay: %6.5f\n",
	       i, segments[i].startGPS.gpsSeconds,
	       segments[i].startGPS.gpsNanoSeconds,
	       segments[i].endGPS.gpsSeconds,
	       segments[i].endGPS.gpsNanoSeconds,
	       segments[i].delay                    );
      fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 2*/

      /* increment tmpTime by segMaxDelay to put some space between segments */
      tmpGPS.gpsSeconds = (INT4) (2 * params->segMaxDelay /
				  params->sampleRate );
      tmpGPS.gpsNanoSeconds = (INT4) ( ((2 * params->segMaxDelay) %
					params->sampleRate) /
				       ((REAL4) params->sampleRate) * 1.0E9 );
      /* increment the number of filled segments */
      output->filled++;
    }		

  /* exit normally */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

/* FillFSeries takes two multiDimData pointers (possibly NULL), one which
 * contains an optimal filter spectrum, the second which contains a sequence of
 * central frequencies and 1/2-widths (in Hz) for frequency regions which need
 * to be  notched from the spectra.  It translates this sequence into a valid
 * COMPLEX8FrequencySeries structure with the appropriate regions "notched out"
 * (multiplied by 10^-35).  This structure is implicit in the ETSearchParams
 * structure and is assumed to be already allocated.
 */
static void
FillFSeries(
	    LALStatus               *status,
	    const multiDimData      *const kernel,
	    const multiDimData      *const notches,
	    ETSearchParams          *const params  ) {
  UINT4                       i,j;        /* iterator(s)                  */
  CHAR                  errstr[500];      /* temporary error string       */
  CHAR                  namestr[500];     /* temporary name string        */
  /* pointers for various places in the ETSearchParams structure */
  ETInitParams         *initParams=params->initParams;
  /* frequency series variables (shortcuts to ETSearchParams frequency data) */
  COMPLEX8Sequence     *kernData, *inputData, *notchData; /* data ptrs    */
  UINT4                 kernLength =      /* length of freq kernel        */
    initParams->segLength / 2 + 1;
                                      /* = kernel->dimensions[0]          */
  REAL8                 kernDeltaF =      /* frequency step for kernel    */
    ((REAL4) initParams->sampleRate) / initParams->segLength;
                                      /* =kernel->range.dFreq.freqStepSize*/
  CHARVector           *tmpFreqUnits;     /* used to parse units string   */
  /* variables to identify individual frequencies */
  COMPLEX8              identity, small, zero;/* constants                */
  REAL8                 minFreq,maxFreq;
  UINT4                 minFreqIdx, maxFreqIdx; /* indices of freq ranges */


  INITSTATUS( status, "FillFSeries", EXTTRIGCONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "%s -- FillFSeries...\n", params->id );
  fflush ( stdout );
#endif /*  DEBUG_EXTTRIG > 2*/

  /**********************
   **********************
   ***  Sanity checks on initParams relevant to LALConditionData
   **********************
   **********************/
    

  /**********************
   **********************
   ***  Allocating memory for <datatype>Sequence; initiating data
   ***  (tmpFreqUnits, fsKernel, fsNotch, filtData)
   **********************
   **********************/
  /* allocate and initialize tmpFreqUnits */
  LALCHARCreateVector( status->statusPtr, &tmpFreqUnits, maxMultiDimUnits );
  CHECKSTATUSPTR( status );

  /* initialize COMPLEX8 values identity, "zero" */
  identity.re = 1.0;
  small.re    = 1E-35;
  identity.im = small.im = zero.re = zero.im = 0.0;
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: filter kernel length = %d\n", params->id,kernLength );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

  /* filter kernel: just make this the identity matrix for now */
  LALCCreateVector( status->statusPtr, &(params->fsKernel->data), kernLength );
  CHECKSTATUSPTR( status );
  kernData = params->fsKernel->data;

  /* input kernel: just make this the identity matrix for now */
  LALCCreateVector( status->statusPtr, &inputData, kernLength );
  CHECKSTATUSPTR( status );

  /* frequency notch: identity through all frequencies for now */
  LALCCreateVector( status->statusPtr, &(params->fsNotch->data), kernLength );
  CHECKSTATUSPTR( status );
  notchData  = params->fsNotch->data;
  for ( i = 0; i < kernLength; i++ ) 
    notchData->data[i] = inputData->data[i] = kernData->data[i]  = identity;

  /**********************
   **********************
   *** Check for valid input kernel pointer; assign data as necessary
   **********************
   **********************/
  if ( kernel ) {
    strncpy( params->fsKernel->name, params->kernel, dbNameLimit );
    params->fsKernel->f0          = kernel->range.dFreq.baseFreq;
    params->fsKernel->deltaF      = kernel->range.dFreq.freqStepSize;
    strncpy( tmpFreqUnits->data, kernel->units, maxMultiDimUnits );
    LALParseUnitString( status->statusPtr,
			&(params->fsKernel->sampleUnits), tmpFreqUnits );
    CHECKSTATUSPTR( status );
    if (kernel->type == complex_8) {
      /* :NOTE: could we just use memcpy to copy over complex8 data? */
      for (i=0; i < kernLength; i++)
	inputData->data[i] = kernel->data.complex8[i];
    } else if (kernel->type == complex_16) {
      for (i=0; i < kernLength; i++) {
	inputData->data[i].re = (REAL4) kernel->data.complex16[i].re;
	inputData->data[i].im = (REAL4) kernel->data.complex16[i].im;
      }
    }
  } else { /* No valid kernel pointer is present */
#if DEBUG_EXTTRIG > 1
    fprintf( stdout, "%s: no kernel frequency series multiDimData\n",
	     params->id );
    fflush ( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
    sprintf( namestr, "%s/* %s_CC_kernel", params->chan1, params->chan2 );
    strncpy( params->fsKernel->name, namestr, LALNameLength - 1 );
    params->fsKernel->f0          = 0;
    params->fsKernel->deltaF      = kernDeltaF;
    params->fsKernel->sampleUnits = lalHertzUnit;
  }

#if DEBUG_EXTTRIG > 3
  fprintf( stdout, "%s: kernel [%s] location: %p,\t notch [%s] location: %p\n",
	   params->id, params->fsKernel->name, params->fsKernel,
	   params->fsNotch->name, params->fsNotch );
    fflush ( stdout );
#endif /*  DEBUG_EXTTRIG > 3*/


  /**********************
   **********************
   *** Check for valid notches pointer; assign data as necessary
   **********************
   **********************/
  if ( notches ) {
    strncpy( params->fsNotch->name, params->notch, dbNameLimit );
    params->fsNotch->f0          = notches->range.dFreq.baseFreq;
    params->fsNotch->deltaF      = notches->range.dFreq.freqStepSize;
    strncpy( tmpFreqUnits->data, notches->units, maxMultiDimUnits );
    LALParseUnitString( status->statusPtr,
			&(params->fsNotch->sampleUnits), tmpFreqUnits );
    CHECKSTATUSPTR( status );
    for( i=0; i < notches->dimensions[0]; i+= 2) {
      if (notches->type == real_4) {
	minFreq = (REAL8) (notches->data.real4[i] - notches->data.real4[i+1]);
	maxFreq = (REAL8) (notches->data.real4[i] + notches->data.real4[i+1]);
      } else if ( notches->type == real_8  ) {
	minFreq = notches->data.real8[i] - notches->data.real8[i+1];
	maxFreq = notches->data.real8[i] + notches->data.real8[i+1];
      } else {
	ABORT( status, CONDITIONDATAH_ETYPE, CONDITIONDATAH_MSGETYPE );
      } /*  if (notches->type == real_4) ...*/

      /* Assign frequency series data indices */
      /* :TODO: Correct these indices for possible heterodyned time series */
      minFreqIdx =  (UINT4) (2 * minFreq * (kernLength - 1) /
			     initParams->sampleRate          ) + 1;
      maxFreqIdx =  (UINT4) (2 * maxFreq * (kernLength - 1) /
			     initParams->sampleRate          ) + 1;  
      /* make sure minFreqIdx and maxFreqIdx are valid freq data indices */
      if ( (minFreqIdx > kernLength) || (maxFreqIdx < 0) )	{
	ABORT( status, CONDITIONDATAH_EFREQ, CONDITIONDATAH_MSGEFREQ );
      }
      if (minFreqIdx < 0) minFreqIdx = 0;
      if (maxFreqIdx > kernLength) maxFreqIdx = kernLength;

      /* travel through the frequency series until the notch is out */
      for( j=minFreqIdx; j < maxFreqIdx; j++ ) notchData->data[j] = small;
    } /*  for( i=0; i < notches->dimensions[0]; i++) ...*/
  } else {
#if DEBUG_EXTTRIG > 1
    fprintf( stdout, "%s: no notch raw sequence multiDimData\n", params->id );
    fflush ( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/
    sprintf( namestr, "%s/* %s_CC_notch", params->chan1, params->chan2 );
    strncpy( params->fsNotch->name, namestr, LALNameLength - 1 );
    params->fsNotch->f0          = 0;
    params->fsNotch->deltaF      = kernDeltaF;
    params->fsNotch->sampleUnits = lalHertzUnit;
  }

#if DEBUG_EXTTRIG > 3
  fprintf( stdout, "%s: kernel [%s] location: %p,\t notch [%s] location: %p\n",
	   params->id, params->fsKernel->name, params->fsKernel,
	   params->fsNotch->name, params->fsNotch );
    fflush ( stdout );
#endif /*  DEBUG_EXTTRIG > 3*/


  /**********************
   **********************
   *** Apply the frequency window to fsNotch
   **********************
   **********************/
  /* :TODO: Correct these indices for possible heterodyned time series */
  minFreqIdx =  (UINT4) (2 * initParams->minFrequency *
			 (kernLength - 1) / initParams->sampleRate) + 1;
  maxFreqIdx =  (UINT4) (2 * initParams->maxFrequency *
			 (kernLength - 1) / initParams->sampleRate) + 1;  
  /* check that minFreqIdx and maxFreqIdx are within the range of fsKernel */
  if ( (minFreqIdx > kernLength) || (maxFreqIdx > kernLength) ||
       (minFreqIdx < 0) || (maxFreqIdx < 0) || (minFreqIdx >= maxFreqIdx) ) {
      ABORT( status, CONDITIONDATAH_EFREQ, CONDITIONDATAH_MSGEFREQ );
  }
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: minFreqIdx = %d; maxFreqIdx = %d\n",
	   params->id, minFreqIdx, maxFreqIdx );
  fflush ( stdout );
#endif /*  DEBUG_EXTTRIG > 1  */
  for (i=0; i < minFreqIdx; i++)          notchData->data[i] = zero;
  for (i=maxFreqIdx; i < kernLength; i++) notchData->data[i] = zero;
#if DEBUG_EXTTRIG > 4
  fprintf( stdout, "%s: fsKernel data:\n", params->id );
  for (i=0; i < params->fsKernel->data->length; ) {
    fprintf( stdout, "row %4d: ", i );
    for (j=0; j < 5; j++) {
      fprintf( stdout, "%10.3g %10.3g ",
	       params->fsKernel->data->data[i].re,
	       params->fsKernel->data->data[i].im );
      i++;
      if ( i == params->fsKernel->data->length ) { break; }
    }
    fprintf( stdout, "\n" );
    fflush( stdout );
  }
  fprintf( stdout, "%s: fsNotch data:\n", params->id );
  for (i=0; i < params->fsNotch->data->length; ) {
    fprintf( stdout, "row %4d: ", i );
    for (j=0; j < 5; j++) {
      fprintf( stdout, "%10.3g %10.3g ",
	       params->fsNotch->data->data[i].re,
	       params->fsNotch->data->data[i].im );
      i++;
      if ( i == params->fsNotch->data->length ) { break; }
    }
    fprintf( stdout, "\n" );
    fflush( stdout );
  }
#endif /*  DEBUG_EXTTRIG > 4*/


  /**********************
   **********************
   *** Finally, multiply fsNotch with the filter kernel
   **********************
   **********************/
  LALCCVectorMultiply( status->statusPtr, kernData, inputData, notchData );
  CHECKSTATUSPTR( status );

#if DEBUG_EXTTRIG > 1
    fprintf( stdout, "%s: kernel [%s] stats: length=%d,\tf0 = %6.2g\tdeltaF = %6.2g\n",
	     params->id, params->fsKernel->name,
	     params->fsKernel->data->length,
	     params->fsKernel->f0, params->fsKernel->deltaF );
    fprintf( stdout, "%s: notch  [%s]  stats: length=%d,\tf0 = %6.2g\tdeltaF = %6.2g\n",
	     params->id, params->fsNotch->name,
	     params->fsNotch->data->length,
	     params->fsNotch->f0, params->fsNotch->deltaF );
    fflush ( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

  /* clean up */
  LALCDestroyVector( status->statusPtr, &inputData );
  CHECKSTATUSPTR( status );

  /* exit normally */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

#ifdef __cplusplus
}
#endif
