#!/usr/bin/env tclshexe
#
# $Id: ldasinspiral.tcl,v 1.10 2003/04/17 02:53:42 duncan Exp $
#
package require LDASJob

;## check the command line arguments for valid args
if { $argc != 4 } {
  puts stderr "usage: $argv0 interferometer channel datatype usertag"
  puts stderr "  interferometer:        IFO code (L1, H1 or H2)"
  puts stderr "  channel:               name of channel (e.g. LSC-AS_Q)"
  puts stderr "  datatype:              frame data type (R, RDS_R_L1)"
  puts stderr "  usertag:               LDAS userTag"
  exit 1
}

;## parse the command line arguments
set ifo [ lindex $argv 0 ]
if { $ifo != "H1" && $ifo != "H2" && $ifo != "L1" } {
  puts stderr "invalid ifo code: must be H1, H2 or L1"
  exit 1
}
set channame [ lindex $argv 1 ]
set datatype [ lindex $argv 2 ]
set usertag  [ lindex $argv 3 ]

;## open the logfile and the error file
set rcsid {$Id: ldasinspiral.tcl,v 1.10 2003/04/17 02:53:42 duncan Exp $}
set rcsid [ string trim $rcsid "\$" ]
set logfilename "ldasinspiral_${ifo}-${channame}_${usertag}"
if { [ catch { open "$logfilename.out" a } logfile ] } {
   puts stderr "error opening output log file $logfilename:\n$logfile"
   exit 1
}
fconfigure $logfile -buffering line
puts $logfile "---------------------------------------------------------------"
puts $logfile "script started at [clock format [clock seconds]]"
puts $logfile $rcsid
if { [ catch { open "$logfilename.err" a } errfile ] } {
   puts stderr "error opening $logfilename.err:\n$errfile"
   exit 1
}
fconfigure $errfile -buffering line

;## read the parameter file
if { [ catch { source "ldasinspiral.par.tcl" } err ] } {
   puts stderr "error sourcing ldasinspiral.par.tcl:\n$err"
   exit 1
} else {
   puts $logfile "read parameters from ldasinspiral.par.tcl"
}

;## override with a sub parameters file, if it exists
if { [ file exists "ldasinspiral.sub.tcl" ] } {
   if { [ catch { source "ldasinspiral.sub.tcl" } err ] } {
      puts stderr "error sourcing ldasinspiral.sub.tcl:\n$err"
      exit 1
   } else {
      puts $logfile "read parameters from ldasinspiral.sub.tcl"
   }
}

;## open the input and output segment files
set segfilename "segment_${usertag}"
if { [ catch { open "$segfilename.list" r } segfile ] } {
   puts stderr "error opening $segfilename.list:\n$segfile"
   exit 1
}
if { [ file exists "$segfilename.out" ] } {
   set now [ clock seconds ]
   file rename "$segfilename.out" "$segfilename.out.$now"
   puts $logfile \
      "$segfilename.out exists, renaming to $segfilename.out.$now"
}
if { [ catch { open "$segfilename.out" w } seglog ] } {
   puts stderr "error opening $segfilename.out:\n$seglog"
   exit 1
}
fconfigure $seglog -buffering line

;## create the frame query and dataconditioning algorithms
set dc_algorithms "schan = slice( fchan,$ovrlap,$chunk,1 ); \
                   output( schan,_,_,,channel ); "

if { $channame == "LSC-AS_Q" } {
   ;## adc channel with resampling in the frameAPI
   ;## no further datacondAPI manipulation needed
   set fquery "adc($channel!resample!${resamp}!)"
   set chalias "fchan"
   set dynrange 69.0
} else {
   ;## convert the channel to a float in datacondAPI
   set dc_algorithms \
      "fchan = float(rchan); $dc_algorithms ; clear(fchan);"
   if { $datatype == "R" } {
      ;## adc channel with resampling in the frameAPI
      set fquery "adc($channel!resample!${resamp}!)"
      set chalias "rchan"
   } else {
      ;## proc channel with resampling in datacondAPI
      set fquery "proc($channel)"
      set chalias "chan"
      set dc_algorithms \
         "rchan = resample( chan,1,${resamp} ); \
         $dc_algorithms ; clear( rchan );"
   }
   ;## create a response function
   set deltaf [ expr ( double($srate) / double($numpoints) ) ]
   set tlen [ expr $numpoints / $srate ]
   set dynrange 1.0
   set dc_algorithms "$dc_algorithms \
                      clear( schan ); \
                      clear( spec ); \
                      p = ramp( $fpoints,0.0,1.0 ); \
                      q = ramp( $fpoints,0.0,0.0 ); \
                      a = float( p ); \
                      b = float( q ); \
                      clear( p ); \
                      clear( q ); \
                      r = complex( a,b ); \
                      resp = fseries( r,0.0,$deltaf,0,0,$tlen,0 ); \
                      clear( r ); \
                      output( resp,_,_,,response );"
}

set filterparams "($numpoints,$numseg,$ovrlap,$flow,$istrunc,$dynrange,$numchisqbins,$maxoverchirp,$rhosq,$chisq,1,$numtmpltexch,$debugflag,$tmpltparams)"

#-----------------------------------------------------------------------------

;## loop over the number of lock segments that we have
while { [ gets $segfile line ] >= 0 } {

   ;## get the parameters for the current data segment
   set doneflag [ lindex $line 0 ]
   set usestart [ lindex $line 1 ]
   set usestop  [ lindex $line 2 ]
   set segnum   [ lindex $line 3 ]

   ;## print out the data segment about to be processes
   puts $logfile \
      "processing $length seconds of data from $usestart to $usestop"
   set startstop "$usestart $usestop"

   ;## construct time-range string, accounting for the fact 
   ;## that LDAS expects us to specify an end time which is 
   ;## 1 second before the actual end of the data we want
   set times \
      "[expr $usestart-$ovrlaptime]-[expr {$usestop+$ovrlaptime-1}]"

   ;## If the job has not been executed yet,  try to run it
   if { ! $doneflag } {

      ;## run the job using the ldasjhob package
      LJrun job1 -manager $manager -log \
      { puts $logfile "$job1(jobid) submitted" } \
      { dataPipeline
         -framequery { { $datatype } { [string index $channel 0] } {} 
         { $times } { $fquery } 
         }
         -aliases { $chalias = $channame; }
         -responsefiles { $response $bankfile }
         -algorithms { $dc_algorithms }
         -datacondtarget wrapper
         -dynlib $dynlib
         -np $nprocs
         -filterparams $filterparams
         -mddapi ligolw     
         -usertag { $usertag }
         -subject { $usertag:$channel:$times }
         -database $db
      }

      if $LJerror {
         puts $logfile "LDAS job failed!"
         catch { puts $logfile $job1(jobReply) }
         if { [ catch {
            puts $errfile "$job1(jobid) $startstop $job1(error)"
         } err ] } {
            puts $errfile "NULL $startstop could not submit job to manager"
         }

         ;## The job failed,  record in the segment log file
         puts $seglog "$doneflag $usestart $usestop $segnum"
      } else {
         puts $logfile \
            "LDAS job $job1(jobid) succeeded.  Reply message from LDAS:"
         puts $logfile $job1(jobReply)

         ;## The job ran successfully,  update the segment log file
         incr doneflag
         puts $seglog "$doneflag $usestart $usestop $segnum"
      }

      LJdelete job1

   } else {
        ;## the job was already run successfully
        puts $logfile "NULL job already completed, ignoring"
        puts $seglog "$doneflag $usestart $usestop $segnum"
   }

   if { [ file exists "sHuTdOwN" ] } {
      puts $logfile "found a shutdown file: cleaning up and exiting"
      while { [ gets $segfile line ] >= 0 } {
         puts $seglog $line
      }
      file delete sHuTdOwN
   }

}

puts $logfile "script finished at [clock format [clock seconds]]"
set now [ clock seconds ]
file rename "$segfilename.list" "$segfilename.list.$now"
puts $logfile "$segfilename.list renamed to $segfilename.list.$now"


close $logfile
close $errfile
close $segfile
close $seglog

exit 0
