/********************************** <lalVerbatim file="StochasticInitSearchHV">
Author: Anderson, W.G., Romano, J.D., and Whelan, J.T.
$Id: SGWBInitSearch.h,v 1.2 2003/09/29 20:11:29 kaice Exp $
**************************************************** </lalVerbatim> */

/******************************* <lalLaTeX file="StochasticInitSearchH">
\section{Header \texttt{StochasticInitSearch.h}}
\label{s:StochasticInitSearch.h}

Header file for a routine that checks the validity of the input 
parameters, initializes them, and then allocates memory for certain
data structures used in the stochastic search.

\subsection*{Synopsis}
\begin{verbatim}
#include "StochasticInitSearch.h"
\end{verbatim}

\subsection*{Error conditions}
\input{StochasticInitSearchHErrTab}

\subsection*{Structures}

None.

*************************************************** </lalLaTeX> */

#ifndef _SGWBINITSEARCH_H
#define _SGWBINITSEARCH_H

#include <lal/LALStdlib.h>
#include "LALWrapperInterface.h"
#include "SGWB.h"

#ifdef  __cplusplus
extern "C" {
#endif

NRCSID (STOCHASTICINITSEARCHH, "$Id: SGWBInitSearch.h,v 1.2 2003/09/29 20:11:29 kaice Exp $");

/******** <lalErrTable file="StochasticInitSearchHErrTab"> ********/
#define STOCHASTICINITSEARCHH_ENULL        1
#define STOCHASTICINITSEARCHH_ENNUL        2
#define STOCHASTICINITSEARCHH_EALOC        3
#define STOCHASTICINITSEARCHH_EARGS        4
#define STOCHASTICINITSEARCHH_EARG0        5
#define STOCHASTICINITSEARCHH_ENSEG        6
#define STOCHASTICINITSEARCHH_ENODENSEG    7
#define STOCHASTICINITSEARCHH_ENPTS        8
#define STOCHASTICINITSEARCHH_EBASE        9
#define STOCHASTICINITSEARCHH_EUWIN       10
#define STOCHASTICINITSEARCHH_ENWIN       11
#define STOCHASTICINITSEARCHH_ENFRQ       12
#define STOCHASTICINITSEARCHH_EF0         13
#define STOCHASTICINITSEARCHH_EDELF       14
#define STOCHASTICINITSEARCHH_EDTSZ       15
#define STOCHASTICINITSEARCHH_EFRES       16
#define STOCHASTICINITSEARCHH_EFRNG       17
#define STOCHASTICINITSEARCHH_EFNUM       18

#define STOCHASTICINITSEARCHH_MSGENULL    "Null pointer"
#define STOCHASTICINITSEARCHH_MSGENNUL    "Non-null pointer"
#define STOCHASTICINITSEARCHH_MSGEALOC    "Memory allocation error"
#define STOCHASTICINITSEARCHH_MSGEARGS    "Wrong number of arguments"
#define STOCHASTICINITSEARCHH_MSGEARG0    "Wrong value for argv[0]"
#define STOCHASTICINITSEARCHH_MSGENSEG    "Number of data segments is zero or negative"
#define STOCHASTICINITSEARCHH_MSGENODENSEG  "Number of segments per nodeis zero or negative"
#define STOCHASTICINITSEARCHH_MSGENPTS    "Number of points in a data segment is zero or negative"
#define STOCHASTICINITSEARCHH_MSGEBASE    "Heterodyning base frequency is negative"
#define STOCHASTICINITSEARCHH_MSGEUWIN    "Unknown window type"
#define STOCHASTICINITSEARCHH_MSGENWIN    "Flat top longer than window"
#define STOCHASTICINITSEARCHH_MSGENFRQ    "Number of frequencies is zero or negative"
#define STOCHASTICINITSEARCHH_MSGEF0      "Start frequency is negative"
#define STOCHASTICINITSEARCHH_MSGEDELF    "Frequency spacing is zero or negative"
#define STOCHASTICINITSEARCHH_MSGEDTSZ    "Nonstandard size of COMPLEX8 and/or UCHAR"
#define STOCHASTICINITSEARCHH_MSGEFRES    "Frequency sampling for output finer than for optimal filter"
#define STOCHASTICINITSEARCHH_MSGEFRNG    "Frequency range for output not fully contained in  optimal filter"
#define STOCHASTICINITSEARCHH_MSGEFNUM    "Number of frequencies in optimal filter is greater than number of data elements"

/******** </lalErrTable> ********/

/*
 * libstochastic.so should be invoked with the following arguments to wrapperAPI
 *
 * libldasstochastic.so should be invoked with the following arguments
 *
 * argv[1]  = numSegments       Number of data segments
 * argv[2]  = numPoints         Number of points in a data segment
 * argv[3]  = nodeNumSegs       Number data segment used per node
 * argv[4]  = baseFreq          Hetrodyning base frequency of time series data
 * argv[5]  = windowType1       Index of window type (0=rect, 1=Hann, etc)
 * argv[6]  = windowTop1        Width of flat top in middle of window
 * argv[7]  = windowType2       Index of window type (0=rect, 1=Hann, etc)
 * argv[8]  = windowTop2        Width of flat top in middle of window
 * argv[9]  = fullLength        Length after zero-padding
 * argv[10]  = numFrequencies    Number of frequencies in optimal filter
 * argv[11] = f0                Start frequency for optimal filter
 * argv[12] = deltaF            Frequency spacing for optimal filter
 * argv[13] = outputNumFreq     Number of frequencies in CC spectrum
 * argv[14] = outputF0          Start frequency for CC spectrum
 * argv[15] = outputDeltaF      Frequency spacing for CC spectrum
 * argv[16] = epochsMatch       Check that time streams are simultaneous
 * argv[17] = siteID1           Identifier for first detector geometry
 * argv[18] = siteID2           Identifier for second detector geometry
 * argv[19] = outputType        Bitmask for DSO outputs 
 *
 */

typedef enum {
  STOCHASTICFILTERPARAM_FILTERPARAMS,
  STOCHASTICFILTERPARAM_NUMSEGMENTS,
  STOCHASTICFILTERPARAM_NUMPOINTS,
  STOCHASTICFILTERPARAM_NODENUMSEGS,
  STOCHASTICFILTERPARAM_BASEFREQ,
  STOCHASTICFILTERPARAM_WINDOWTYPE1,
  STOCHASTICFILTERPARAM_WINDOWTOP1, 
  STOCHASTICFILTERPARAM_WINDOWTYPE2,
  STOCHASTICFILTERPARAM_WINDOWTOP2,
  STOCHASTICFILTERPARAM_FULLLENGTH,
  STOCHASTICFILTERPARAM_NUMFREQUENCIES,
  STOCHASTICFILTERPARAM_F0,
  STOCHASTICFILTERPARAM_DELTAF,
  STOCHASTICFILTERPARAM_OUTPUTNUMFREQ,
  STOCHASTICFILTERPARAM_OUTPUTF0,
  STOCHASTICFILTERPARAM_OUTPUTDELTAF,
  STOCHASTICFILTERPARAM_EPOCHSMATCH,
  STOCHASTICFILTERPARAM_SITEID1,
  STOCHASTICFILTERPARAM_SITEID2,
  STOCHASTICFILTERPARAM_OUTPUTTYPE,
  STOCHASTICFILTERPARAM_FREQMASK,
  STOCHASTICFILTERPARAM_NPARAMS
}
StochasticFilterParam;

#define MAX_STRING_LENGTH 255
#define MAX_TOKS 5000
#define DELIMITERS " ,"

  void LALParseMaskParams(  LALStatus *status,
			    CHAR* freqMask, 
			    CHAR*** maskElements,
			    INT4 *numElements,
			    StochasticSearchParams *params);

#ifdef  __cplusplus
}
#endif /* C++ protection */

#endif /* _SGWBINITSEARCH_H  */


/******************************* <lalLaTeX file="StochasticInitSearchH">
  
\vfill{\footnotesize\input{StochasticInitSearchHV}}

\newpage\input{StochasticInitSearchC}

*************************************************** </lalLaTeX> */
