/**** <lalVerbatim file="CalibrationCV" >
 * $Id: Calibration.c,v 1.3 2002/10/24 21:14:03 jolien Exp $
 **** </lalVerbatim> */

/**** <lalLaTeX>
 *
 * \subsection{Module \texttt{Calibration.c}}
 * \label{ss:Calibration.c}
 *
 * This module contains code used to extract calibration information contained
 * in a \texttt{LALSearchInput} structure, and to construct a response
 * (or transfer) function.  This is supposed to provide a high-level interface
 * for DSO authors to obtain a response function in the desired form.
 *
 * \subsection*{Prototypes}
 * \input{CalibrationCP}
 *
 * The routine \texttt{LALExtractResponse()} extracts the necessary
 * calibration information from the \texttt{LALSearchInput} structure
 * and constructs a response function (as a frequency series) from this
 * information.  If the fourth argument is non-\texttt{NULL} then this
 * string specifies the detector (H1, H2, L1, etc.) for which the calibration
 * is required; if this argument is \texttt{NULL}, the first calibration
 * information found will in the \texttt{LALSearchInput} will be used.
 *
 * Certain fields of the output should be set before this routine is
 * called.  In particular: 1. The epoch field of the frequency series
 * should be set to the correct epoch so that the routine can generate
 * a response function tailored to that time (accounting for calibration
 * drifts).  2. The units of the response function should be set to
 * be either strain-per-count (for a response function) or count-per-strain
 * (for a transfer function); the routine will then return either the response
 * function or its inverse depending on the specified units.  Furthermore,
 * the power-of-ten field of the units is examined to scale the response
 * function accordingly.  3. The data vector should be allocated to the
 * required length and the frequency step size should be set to the required
 * value so that the routine can interpolate the response function to the
 * required frequencies.
 *
 * Note that the \texttt{LALSearchInput} must contain the appropriate
 * calibration \texttt{multiDimData} sequences, which must be input to
 * the \texttt{wrapperAPI}.
 *
 * \vfill{\footnotesize\input{CalibrationCV}}
 *
 **** </lalLaTeX> */
#include <string.h>
#include <lal/LALStdlib.h>
#include <lal/AVFactories.h>
#include <lal/Calibration.h>
#include <ExtractSeries.h>

NRCSID( CALIBRATIONC, "$Id: Calibration.c,v 1.3 2002/10/24 21:14:03 jolien Exp $" );

#define RESPONSE_REGEX "CAL-RESPONSE"
#define CAV_GAIN_REGEX "CAL-CAV_GAIN"
#define CAV_FAC_REGEX "CAL-CAV_FAC"
#define OLOOP_FAC_REGEX "CAL-OLOOP_FAC"

/* <lalVerbatim file="CalibrationCP"> */
void
LALExtractResponse(
    LALStatus               *status,
    COMPLEX8FrequencySeries *output,
    LALSearchInput          *input,
    const CHAR              *ifo
    )
{ /* </lalVerbatim> */
  const LALUnit strainPerCount = {0,{0,0,0,0,0,1,-1},{0,0,0,0,0,0,0}};
  CHAR regex[64];
  CHAR regexhead[64];
  COMPLEX8Vector abvec;
  COMPLEX8Vector avec;
  COMPLEX8FrequencySeries R0;
  COMPLEX8FrequencySeries C0;
  COMPLEX8TimeSeries ab;
  COMPLEX8TimeSeries a;
  CalibrationFunctions    calfuncs;
  CalibrationUpdateParams calfacts;
  multiDimData *R0mdd = NULL;
  multiDimData *C0mdd = NULL;
  multiDimData *amdd  = NULL;
  multiDimData *abmdd = NULL;

  INITSTATUS( status, "LALExtractResponse", CALIBRATIONC );
  ATTATCHSTATUSPTR( status );

  ASSERT( input, status, EXTRACTSERIES_ENULL, EXTRACTSERIES_MSGENULL );
  ASSERT( output, status, EXTRACTSERIES_ENULL, EXTRACTSERIES_MSGENULL );

  strncpy( regexhead, ifo ? ifo : "", sizeof( regexhead ) - 1 );
  strncat( regexhead, ifo ? ".*:" : "",
      sizeof( regexhead ) - strlen( regexhead ) - 1 );
  strncpy( regex, regexhead, sizeof( regex ) - 1 );
  strncat( regex, RESPONSE_REGEX, sizeof( regex ) - strlen( regex ) - 1 );
  LALFindSequenceByName( status->statusPtr, &R0mdd, input, regex );
  CHECKSTATUSPTR( status );
  strncpy( regex, regexhead, sizeof( regex ) - 1 );
  strncat( regex, CAV_GAIN_REGEX, sizeof( regex ) - strlen( regex ) - 1 );
  LALFindSequenceByName( status->statusPtr, &C0mdd, input, regex );
  CHECKSTATUSPTR( status );
  strncpy( regex, regexhead, sizeof( regex ) - 1 );
  strncat( regex, OLOOP_FAC_REGEX, sizeof( regex ) - strlen( regex ) - 1 );
  LALFindSequenceByName( status->statusPtr, &abmdd, input, regex );
  CHECKSTATUSPTR( status );
  strncpy( regex, regexhead, sizeof( regex ) - 1 );
  strncat( regex, CAV_FAC_REGEX, sizeof( regex ) - strlen( regex ) - 1 );
  LALFindSequenceByName( status->statusPtr, &amdd, input, regex );
  CHECKSTATUSPTR( status );

  R0.data    = NULL; /* extract will create vector */
  C0.data    = NULL; /* extract will create vector */
  abvec.data = NULL; /* extract will copy pointer  */
  avec.data  = NULL; /* extract will copy pointer  */
  ab.data = &abvec;
  a.data  = &avec;
  calfuncs.responseFunction = &R0;
  calfuncs.sensingFunction  = &C0;
  calfacts.openLoopFactor   = &ab;
  calfacts.sensingFactor    = &a;
  calfacts.epoch = output->epoch; /* specify correct epoch for update */

  LALExtractCOMPLEX8FrequencySeries( status->statusPtr, &R0, R0mdd );
  CHECKSTATUSPTR( status );
  LALExtractCOMPLEX8FrequencySeries( status->statusPtr, &C0, C0mdd );
  CHECKSTATUSPTR( status );
  LALExtractCOMPLEX8TimeSeries( status->statusPtr, &ab, abmdd );
  CHECKSTATUSPTR( status );
  LALExtractCOMPLEX8TimeSeries( status->statusPtr, &a, amdd );
  CHECKSTATUSPTR( status );

  /* should be able to update into the same functions... */
  calfuncs.responseFunction->sampleUnits = strainPerCount;
  LALUpdateCalibration( status->statusPtr, &calfuncs, &calfuncs, &calfacts );
  BEGINFAIL( status )
  {
    if ( status->statusPtr->statusCode == CALIBRATIONH_EZERO )
    {
      TRY( LALResponseConvert( status->statusPtr, output,
            calfuncs.responseFunction ), status );
    }
    TRY( LALCDestroyVector( status->statusPtr, &R0.data ), status );
    TRY( LALCDestroyVector( status->statusPtr, &C0.data ), status );
  }
  ENDFAIL( status );

  /* now convert response to get output, hardwire units */
  LALResponseConvert( status->statusPtr, output, calfuncs.responseFunction );
  CHECKSTATUSPTR( status );

  /* free up allocated vectors */
  TRY( LALCDestroyVector( status->statusPtr, &R0.data ), status );
  TRY( LALCDestroyVector( status->statusPtr, &C0.data ), status );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}
